# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from dataclasses import dataclass

from someipy._internal._common.endpoint import Endpoint
from someipy._internal.transport_layer_protocol import TransportLayerProtocol


@dataclass
class ServiceInstance:
    """This class aggregates data from entries and options and provides a compact interface instead of loose SD entries and options"""

    service_id: int
    instance_id: int
    major_version: int
    minor_version: int
    ttl: int
    endpoint: Endpoint
    protocols: frozenset[TransportLayerProtocol]
    timestamp: float

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ServiceInstance):
            return False

        # Ignore the timestamp in equality comparison
        return (
            self.service_id == other.service_id
            and self.instance_id == other.instance_id
            and self.major_version == other.major_version
            and self.minor_version == other.minor_version
            and self.ttl == other.ttl
            and self.endpoint == other.endpoint
            and self.protocols == other.protocols
        )



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
from someipy._internal._common.endpoint import Endpoint
from someipy.service import Event, EventGroup, Method
from someipy._internal.transport_layer_protocol import TransportLayerProtocol


class ServiceToOffer:
    def __init__(
        self,
        client_writer_id: int,
        instance_id: int,
        service_id: int,
        major_version: int,
        minor_version: int,
        offer_ttl_seconds: int,
        cyclic_offer_delay_ms: int,
        endpoint: Endpoint,
        methods: List[Method],
        eventgroups: List[EventGroup],
    ):
        self.client_writer_id = client_writer_id
        self.instance_id = instance_id
        self.service_id = service_id
        self.major_version = major_version
        self.minor_version = minor_version
        self.offer_ttl_seconds = offer_ttl_seconds
        self.cyclic_offer_delay_ms = cyclic_offer_delay_ms
        self.endpoint = endpoint
        self.methods = methods
        self.eventgroups = eventgroups
        self.last_offer_time = None  # Placeholder for last offer time

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ServiceToOffer):
            return False
        return (
            self.client_writer_id == other.client_writer_id
            and self.instance_id == other.instance_id
            and self.service_id == other.service_id
            and self.major_version == other.major_version
            and self.minor_version == other.minor_version
            and self.offer_ttl_seconds == other.offer_ttl_seconds
            and self.cyclic_offer_delay_ms == other.cyclic_offer_delay_ms
            and self.endpoint == other.endpoint
            and self.methods == other.methods
            and self.eventgroups == other.eventgroups
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.client_writer_id,
                self.instance_id,
                self.service_id,
                self.major_version,
                self.minor_version,
                self.offer_ttl_seconds,
                self.cyclic_offer_delay_ms,
                self.endpoint,
                tuple(self.methods),  # Convert list to tuple for hashing
                tuple(self.eventgroups),  # Convert list to tuple for hashing
            )
        )

    @property
    def has_udp(self) -> bool:
        has_method_udp = any(
            method.protocol == TransportLayerProtocol.UDP for method in self.methods
        )

        events = [event for eg in self.eventgroups for event in eg.events]
        has_event_udp = any(
            event.protocol == TransportLayerProtocol.UDP for event in events
        )
        return has_method_udp or has_event_udp

    @property
    def has_tcp(self) -> bool:
        has_method_tcp = any(
            method.protocol == TransportLayerProtocol.TCP for method in self.methods
        )

        events = [event for eg in self.eventgroups for event in eg.events]
        has_event_tcp = any(
            event.protocol == TransportLayerProtocol.TCP for event in events
        )
        return has_method_tcp or has_event_tcp

    @property
    def eventgroup_ids(self) -> List[int]:
        """Returns a list of all eventgroup IDs from the stored eventgroups"""
        return [eventgroup.id for eventgroup in self.eventgroups]

    @property
    def protocols(self) -> frozenset[TransportLayerProtocol]:
        result = set()
        if self.has_udp:
            result.add(TransportLayerProtocol.UDP)
        if self.has_tcp:
            result.add(TransportLayerProtocol.TCP)
        return frozenset(result)


class OfferServiceStorage:
    def __init__(self):
        self._services: List[ServiceToOffer] = []

    def add_service(self, service: ServiceToOffer) -> bool:
        """Add a new ServiceToOffer to the storage if it doesn't already exist.
        Returns True if the service was added, False if it was a duplicate."""
        if service in self._services:
            return False
        self._services.append(service)
        return True

    def remove_service(self, service: ServiceToOffer) -> bool:
        """Remove a specific ServiceToOffer from the storage.
        Returns True if the service was found and removed, False otherwise."""
        try:
            self._services.remove(service)
            return True
        except ValueError:
            return False

    def remove_client(self, client_writer_id: int) -> int:
        """Remove all services with the given client_writer_id.
        Returns the number of services removed."""
        initial_count = len(self._services)
        self._services = [
            service
            for service in self._services
            if service.client_writer_id != client_writer_id
        ]
        return initial_count - len(self._services)

    def services_by_cyclic_offer_delay(
        self, cyclic_offer_delay_ms: int
    ) -> List[ServiceToOffer]:
        """Return all services with the specified cyclic_offer_delay_ms"""
        return [
            service
            for service in self._services
            if service.cyclic_offer_delay_ms == cyclic_offer_delay_ms
        ]

    def get_all_services(self) -> List[ServiceToOffer]:
        """Return all services in the storage"""
        return self._services.copy()

    def clear(self) -> None:
        """Remove all services from the storage"""
        self._services.clear()

    def __len__(self) -> int:
        return len(self._services)

    @property
    def cyclic_offer_delays(self) -> List[int]:
        """Returns a list of all unique cyclic_offer_delay_ms values from stored services"""
        return sorted({service.cyclic_offer_delay_ms for service in self._services})



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from someipy._internal.logging import get_logger

_logger_name = "tcp_connection"


class TcpConnection:
    def __init__(self, remote_ip: str, remote_port: int):
        self.remote_ip = remote_ip
        self.remote_port = remote_port
        self.reader = None
        self.writer = None

    async def connect(self, src_ip: str, src_port: int):
        local_addr = (src_ip, src_port)
        self.reader, self.writer = await asyncio.open_connection(
            self.remote_ip, self.remote_port, local_addr=local_addr
        )
        get_logger(_logger_name).debug(
            f"Connected to {self.remote_ip}:{self.remote_port}"
        )

    def is_open(self):
        if self.writer is None or self.writer.is_closing():
            return False
        return not self.reader.at_eof()

    async def close(self):
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()
            self.writer = None
            get_logger(_logger_name).debug(
                f"Connection to {self.remote_ip}:{self.remote_port} closed"
            )



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ipaddress
from typing import NamedTuple, Union


class Endpoint(NamedTuple):
    """Represents a network endpoint with IP address and port."""

    ip: Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
    port: int

    @property
    def is_ipv4(self):
        return isinstance(self.ip, ipaddress.IPv4Address)

    def __str__(self):
        if isinstance(self.ip, ipaddress.IPv6Address):
            return f"[{self.ip}]:{self.port}"
        return f"{self.ip}:{self.port}"



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from dataclasses import dataclass

from someipy._internal._sd.entries.sd_entry import SdEntry, SdEntryType


@dataclass
class SubscribeAckEventGroupEntry(SdEntry):
    service_id: int
    instance_id: int
    major_version: int
    minor_version: int
    ttl: int
    eventgroup_id: int
    counter: int

    def __init__(
        self,
        service_id: int,
        instance_id: int,
        major_version: int,
        ttl: int,
        eventgroup_id: int,
        counter: int,
    ):
        super().__init__(SdEntryType.SUBSCRIBE_EVENT_GROUP_ACK)
        self.service_id = service_id
        self.instance_id = instance_id
        self.major_version = major_version
        self.ttl = ttl
        self.eventgroup_id = eventgroup_id
        self.counter = counter



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum


# PRS_SOMEIP_00191
class ReturnCode(Enum):
    E_OK = 0x00  # No error occurred
    E_NOT_OK = 0x01  # An unspecified error occurred
    E_UNKNOWN_SERVICE = 0x02  # The requested Service ID is unknown.
    E_UNKNOWN_METHOD = 0x03  # The requested Method ID is unknown.
    E_NOT_READY = 0x04  # Service ID and Method ID are known. Application not running.
    E_NOT_REACHABLE = (
        0x05  # System running the service is not reachable (internal error code only).
    )
    E_TIMEOUT = 0x06  # A timeout occurred (internal error code only).
    E_WRONG_PROTOCOL_VERSION = 0x07  # Version of SOME/IP protocol not supported
    E_WRONG_INTERFACE_VERSION = 0x08  # Interface version mismatch
    E_MALFORMED_MESSAGE = (
        0x09  # Deserialization error, so that payload cannot be de-serialized.
    )
    E_WRONG_MESSAGE_TYPE = 0x0A  # An unexpected message type was received (e.g. REQUEST_NO_RETURN for a method defined as REQUEST).



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import argparse
import asyncio
import base64
from dataclasses import dataclass
import functools
import json
import logging
import os
import struct
import sys
import ipaddress
import time
from typing import Any, Dict, List, Set, Tuple, Union

import someipy
from someipy._internal._common.endpoint import Endpoint
from someipy._internal._daemon.daemon_server_client import (
    ClientMessageEventArgs,
    DaemonServerClient,
)
from someipy._internal._daemon.subscription import Subscription
from someipy._internal._daemon.subscription_storage import SubscriptionStorage
from someipy._internal._sd.deserialization.sd_deserialization import (
    deserialize_sd_message,
    is_sd_message,
)
from someipy._internal._sd.deserialization.sd_serialization import serialize_sd_message
from someipy._internal._sd.entries.offer_service_entry import OfferServiceEntry
from someipy._internal._sd.entries.stop_subscribe_eventgroup_entry import (
    StopSubscribeEventGroupEntry,
)
from someipy._internal._sd.entries.subscribe_eventgroup_entry import (
    SubscribeEventGroupEntry,
)
from someipy._internal._sd.options.endpoint import IpV4EndpointOption
from someipy._internal._sd.sd_message import SdMessage
from someipy._internal._sd.sd_message_creator import (
    create_offer_service_message,
    create_stop_offer_service_message,
)
from someipy._internal._sd.service_instance import ServiceInstance
from someipy._internal.message_types import MessageType
from someipy._internal.someip_endpoint import (
    TCPClientSomeipEndpoint,
    TCPSomeipEndpoint,
)
from someipy._internal.someip_endpoint_factory import SomeipEndpointFactory
from someipy._internal.someip_endpoint_storage import SomeipEndpointStorage
from someipy._internal.someip_message import SomeIpMessage
from someipy._internal.transport_layer_protocol import TransportLayerProtocol
from someipy._internal.session_handler import SessionHandler
from someipy._internal.simple_timer import SimplePeriodicTimer
from someipy._internal.someip_header import SomeIpHeader
from someipy._internal.someip_sd_builder import (
    build_subscribe_eventgroup_ack_entry,
    build_subscribe_eventgroup_ack_sd_header,
)
from someipy._internal.someip_sd_extractors import (
    extract_subscribe_ack_eventgroup_entries,
    extract_subscribe_entries,
    extract_subscribe_nack_eventgroup_entries,
)
from someipy._internal.someip_sd_header import (
    SdEntryType,
    SdEventGroupEntry,
    SdService,
    SdService2,
    SdServiceWithTimestamp,
    SdSubscription,
    SomeIpSdHeader,
)
from someipy._internal.subscribers import EventGroupSubscriber, Subscribers
from someipy._internal._daemon.uds_messages import (
    InboundCallMethodRequest,
    InboundCallMethodResponse,
    FindServiceRequest,
    FindServiceResponse,
    OfferServiceRequest,
    OutboundCallMethodRequest,
    OutboundCallMethodResponse,
    ReceivedEvent,
    SendEventRequest,
    StopOfferServiceRequest,
    StopSubscribeEventGroupRequest,
    SubscribeEventGroupRequest,
    create_uds_message,
)
from someipy._internal.utils import (
    DatagramAdapter,
    create_rcv_multicast_socket,
    create_rcv_broadcast_socket,
    create_udp_socket,
)
from someipy._internal._daemon.offer_service_storage import (
    OfferServiceStorage,
    ServiceToOffer,
)
from someipy.service import Event, Method, EventGroup
from someipy._internal._daemon.daemon_server import (
    ClientConnectedEventArgs,
    DaemonServer,
)


DEFAULT_SOCKET_PATH = "/tmp/someipyd.sock"
DEFAULT_CONFIG_FILE = "someipyd.json"
DEFAULT_SD_ADDRESS = "224.224.224.245"
DEFAULT_INTERFACE_IP = "127.0.0.1"
DEFAULT_SD_PORT = 30490
DEFAULT_TCP_PORT = 30500


@dataclass
class MethodCall:
    service_id: int
    method_id: int
    client_id: int
    session_id: int
    src_ip: str
    src_port: int

    def __hash__(self):
        return hash(
            (
                self.service_id,
                self.method_id,
                self.client_id,
                self.session_id,
                self.src_ip,
                self.src_port,
            )
        )


class SomeipDaemon:

    def __init__(
        self,
        server: DaemonServer,
        endpoint_factory: SomeipEndpointFactory,
        config: dict = None,
        logger: logging.Logger = None,
    ):

        self.config = config
        self.logger = logger
        self._endpoint_factory = endpoint_factory

        self.sd_address = self.config.get("sd_address", DEFAULT_SD_ADDRESS)
        self.sd_port = self.config.get("sd_port", DEFAULT_SD_PORT)
        self.interface = self.config.get("interface", DEFAULT_INTERFACE_IP)

        self._server = server

        self._sd_socket_mcast = None
        self._sd_socket_ucast = None
        self._mcast_transport = None
        self._ucast_transport = None

        # Services offered by other ECUs
        self._found_services: List[ServiceInstance] = []

        # Services offered by this daemon
        self._services_to_offer = OfferServiceStorage()
        self._offer_timers: Dict[int, SimplePeriodicTimer] = {}

        # Active subscriptions to services offered by this daemon
        self._service_subscribers: Dict[ServiceToOffer, Subscribers] = {}

        # Subscriptions requested by local clients
        self._requested_subscriptions = SubscriptionStorage()

        self._pending_subscriptions: Set[Subscription] = set()
        self._active_subscriptions: Set[Subscription] = set()

        self._mcast_session_handler = SessionHandler()
        self._unicast_session_handler = SessionHandler()

        # Qeueues and tasks stored by id of asyncio.StreamWriter
        self._tx_queues: Dict[int, asyncio.Queue] = {}
        self._tx_tasks: Dict[int, asyncio.Task] = {}

        self._someip_server_endpoints = SomeipEndpointStorage()
        self._someip_client_endpoints = SomeipEndpointStorage()

        self._ttl_task: asyncio.Task = None

        self._issued_method_calls: Dict[MethodCall, int] = {}

    async def new_client_connected(
        self, sender: object, event_args: ClientConnectedEventArgs
    ):
        self.logger.info(f"New client connected: {event_args.client.id}")

        # Create the tx_queue and tx_task for the client using the writers id
        self._tx_queues[event_args.client.id] = asyncio.Queue()
        self._tx_tasks[event_args.client.id] = asyncio.create_task(
            self.tx_task(event_args.client)
        )

        event_args.client.message_received += self.handle_client_message

    async def client_disconnected(
        self, sender: object, event_args: ClientConnectedEventArgs
    ):
        self.logger.info(f"Client disconnected: {event_args.client.id}")

        client_id = event_args.client.id

        event_args.client.message_received -= self.handle_client_message

        # Remove all subscriptions for the client
        self._requested_subscriptions.remove_client(client_id)

        # Clean up the transmission task for the client. This will also clean up the transmission queue
        tx_task = self._tx_tasks.pop(client_id, None)
        if tx_task and not tx_task.cancelled():
            tx_task.cancel()
            try:
                await tx_task
            except asyncio.CancelledError:
                pass

        self._services_to_offer.remove_client(client_id)
        self._cleanup_unused_timers()

        client_endpoints = self._someip_server_endpoints.get_endpoints(client_id)
        if client_endpoints is not None:
            for endpoint in client_endpoints:
                self.logger.debug(
                    f"Closing endpoint {endpoint.dst_ip()}:{endpoint.dst_port()} for client {client_id}"
                )
                endpoint.shutdown()
                self._someip_server_endpoints.remove_endpoint(client_id, endpoint)

        self.logger.debug(f"Client disconnected")

    async def _check_services_ttl_task(self):
        try:
            while True:
                await asyncio.sleep(0.1)
                self._check_service_ttl_task_impl()

        except asyncio.CancelledError:
            # Task was cancelled - exit cleanly
            self.logger.debug("Service TTL checker task cancelled")
        except Exception as e:
            self.logger.error(f"Error in service TTL checker task: {e}")
        pass

    def _check_service_ttl_task_impl(self):

        self._cleanup_obsolete_pending_subscriptions()
        self._cleanup_active_subscriptions()

        count_before = len(self._found_services)

        current_time = time.time()

        # Process timeouts and filter services in one operation
        self._found_services = [
            service
            for service in self._found_services
            if not (current_time - service.timestamp > service.ttl)
        ]

        count_after = len(self._found_services)
        if count_before != count_after:
            self.logger.info(
                f"Removed {count_before - count_after} timed out services. Remaining: {count_after}"
            )

    def _someip_message_callback(
        self,
        message: SomeIpMessage,
        src_addr: Tuple[str, int],
        dst_addr: Tuple[str, int],
        protocol: TransportLayerProtocol,
    ) -> None:
        self.logger.debug(
            f"Received SOME/IP message from {src_addr} to {dst_addr} with protocol {protocol}"
        )

        header = message.header
        if MessageType(header.message_type) == MessageType.REQUEST:
            service_id = header.service_id
            method_id = header.method_id

            for service in self._services_to_offer.get_all_services():

                if (
                    service.service_id == service_id
                    and str(service.endpoint.ip) == dst_addr[0]
                    and service.endpoint.port == dst_addr[1]
                ):
                    self.logger.debug(f"Found matching service {service.service_id}")
                    for method in service.methods:
                        if method.id == method_id:
                            self.logger.debug(
                                f"Found matching method id {method.id} for service {service_id}"
                            )

                            payload_encoded = base64.b64encode(message.payload).decode(
                                "utf-8"
                            )

                            call_method_request = create_uds_message(
                                InboundCallMethodRequest,
                                service_id=service_id,
                                instance_id=service.instance_id,
                                method_id=method_id,
                                client_id=header.client_id,
                                session_id=header.session_id,
                                protocol_version=header.protocol_version,
                                interface_version=header.interface_version,
                                major_version=service.major_version,
                                minor_version=service.minor_version,
                                message_type=header.message_type,
                                src_endpoint_ip=src_addr[0],
                                src_endpoint_port=src_addr[1],
                                protocol=protocol.value,
                                payload=payload_encoded,
                            )

                            tx_queue = self._tx_queues.get(service.client_writer_id)
                            if tx_queue:
                                tx_queue.put_nowait(
                                    self.prepare_message(call_method_request)
                                )

        elif MessageType(header.message_type) == MessageType.RESPONSE:
            issued_call = MethodCall(
                service_id=header.service_id,
                method_id=header.method_id,
                client_id=header.client_id,
                session_id=header.session_id,
                src_ip=dst_addr[0],
                src_port=dst_addr[1],
            )

            if not issued_call in self._issued_method_calls:
                self.logger.debug(f"Received response for unknown method call.")
                return

            writer_id = self._issued_method_calls[issued_call]
            del self._issued_method_calls[issued_call]
            payload_encoded = base64.b64encode(message.payload).decode("utf-8")

            message = create_uds_message(
                OutboundCallMethodResponse,
                service_id=header.service_id,
                method_id=header.method_id,
                client_id=header.client_id,
                session_id=header.session_id,
                return_code=header.return_code,
                dst_endpoint_ip=dst_addr[0],
                dst_endpoint_port=dst_addr[1],
                payload=payload_encoded,
            )

            self._tx_queues[writer_id].put_nowait(self.prepare_message(message))

        elif MessageType(header.message_type) == MessageType.NOTIFICATION:
            if header.return_code == 0x00 and header.client_id == 0x00:
                self.logger.debug(
                    f"Received notification for service {header.service_id}, event {header.method_id}"
                )

                event_id = header.method_id

                event_msg = create_uds_message(
                    ReceivedEvent,
                    service_id=header.service_id,
                    event_id=event_id,
                    src_endpoint_ip=src_addr[0],
                    src_endpoint_port=src_addr[1],
                    payload=base64.b64encode(message.payload).decode("utf-8"),
                )

                for active_subscription in self._active_subscriptions:
                    event_ids = [i.id for i in active_subscription.eventgroup.events]
                    self.logger.debug("check event ids: %s", event_ids)
                    if (
                        event_id in event_ids
                        and str(active_subscription.server_endpoint.ip) == src_addr[0]
                        and active_subscription.server_endpoint.port == src_addr[1]
                        and active_subscription.service_id == header.service_id
                    ):
                        self.logger.debug(
                            f"Found active subscription for service {header.service_id}, "
                            f"instance {active_subscription.instance_id}, "
                            f"eventgroup {active_subscription.eventgroup.id}, "
                        )

                        for (
                            requested_subscription
                        ) in self._requested_subscriptions.subscriptions:
                            if (
                                requested_subscription.service_id
                                == active_subscription.service_id
                                and requested_subscription.instance_id
                                == active_subscription.instance_id
                                and requested_subscription.major_version
                                == active_subscription.major_version
                                and requested_subscription.eventgroup
                                == active_subscription.eventgroup
                                and requested_subscription.client_endpoint.ip
                                == active_subscription.client_endpoint.ip
                                and requested_subscription.client_endpoint.port
                                == active_subscription.client_endpoint.port
                            ):

                                client_ids = (
                                    self._requested_subscriptions.get_client_ids(
                                        requested_subscription
                                    )
                                )

                                for client_id in client_ids:
                                    tx_queue = self._tx_queues.get(client_id)
                                    if tx_queue:
                                        tx_queue.put_nowait(
                                            self.prepare_message(event_msg)
                                        )

    def _cleanup_active_subscriptions(self):
        current_time = time.time()
        subscriptions_to_remove = []
        for subscription in self._active_subscriptions:
            if (
                current_time - subscription.timestamp_last_update
                > subscription.ttl_seconds
            ):
                subscriptions_to_remove.append(subscription)

        for subscription in subscriptions_to_remove:
            self._active_subscriptions.remove(subscription)
            self.logger.debug(
                f"Removed active subscription for service {subscription.service_id}, "
                f"instance {subscription.instance_id}, eventgroup {subscription.eventgroup_id}"
            )

    def _cleanup_obsolete_pending_subscriptions(self):
        current_time = time.time()
        subscriptions_to_remove = []
        for subscription in self._pending_subscriptions:
            if current_time - subscription.timestamp_last_update > 10.0:
                subscriptions_to_remove.append(subscription)

        for subscription in subscriptions_to_remove:
            self._pending_subscriptions.remove(subscription)

    def _cleanup_unused_timers(self):
        timers_to_stop = []
        for interval in self._offer_timers.keys():
            if (
                len(self._services_to_offer.services_by_cyclic_offer_delay(interval))
                == 0
            ):
                timers_to_stop.append(interval)
        for interval in timers_to_stop:
            self.logger.debug(f"Stopping offer timer for {interval}ms")
            self._offer_timers[interval].stop()
            del self._offer_timers[interval]

    def _close_unused_endpoints(self):
        endpoints_to_close = []

        # Loop through all endpoints. For each endpoint loop through offered services and check if the endpoint is used by any service
        for endpoint in self._someip_server_endpoints:
            endpoint_used = False
            for service in self._services_to_offer.services:
                if (
                    service.endpoint.ip == endpoint.ip()
                    and service.endpoint_port == endpoint.port()
                ):
                    endpoint_used = True
                    break

            if not endpoint_used:
                endpoints_to_close.append(endpoint)

    async def tx_task(self, client: DaemonServerClient):
        tx_queue = self._tx_queues[client.id]
        try:
            while True:
                try:
                    # Wait on queue with timeout
                    data = await asyncio.wait_for(tx_queue.get(), timeout=0.2)

                    try:
                        # Send the data
                        await client.send(data)
                        tx_queue.task_done()
                    except ConnectionError as e:
                        self.logger.error(f"Error sending data in tx task: {e}")
                        break

                except asyncio.TimeoutError:
                    # Periodic timeout for cancellation check
                    continue

        except asyncio.CancelledError:
            self.logger.debug(f"TX task for writer {client.id} cancelled")
            # Perform cleanup here
            try:
                await client.close()
            except Exception as e:
                self.logger.error(f"Error closing writer: {e}")
        finally:
            # Always clean up the queue
            self._tx_queues.pop(client.id, None)
            self.logger.debug(f"TX task for writer {client.id} finished")

    async def handle_client_message(
        self, sender: object, event_args: ClientMessageEventArgs
    ):
        client_id = event_args.client.id

        message = event_args.message
        message_type = event_args.message.get("type")
        self.logger.debug(f"Received message type: {message_type}")

        message_handlers = {
            OfferServiceRequest.__name__: self._handle_offer_service_request,
            StopOfferServiceRequest.__name__: self._handle_stop_offer_service_request,
            InboundCallMethodResponse.__name__: self._handle_inbound_call_method_response,
            FindServiceRequest.__name__: self._handle_find_service_request,
            OutboundCallMethodRequest.__name__: self._handle_outbound_call_method_request,
            SendEventRequest.__name__: self._handle_send_event_request,
            SubscribeEventGroupRequest.__name__: self._handle_subscribe_eventgroup_request,
            StopSubscribeEventGroupRequest.__name__: self._handle_stop_subscribe_eventgroup_request,
        }

        if message_type in message_handlers:
            handler = message_handlers[message_type]

            if asyncio.iscoroutinefunction(handler):
                await handler(message, client_id)
                return
            else:
                handler(message, client_id)
                return
        else:
            self.logger.warning(
                f"Received unknown message type: {message_type}. Message: {message}"
            )

    async def _handle_subscribe_eventgroup_request(
        self, message: SubscribeEventGroupRequest, client_id: int
    ):

        protocols = []
        if message["udp"]:
            protocols.append(TransportLayerProtocol.UDP)

            if not self._someip_client_endpoints.has_endpoint(
                message["client_endpoint_ip"],
                message["client_endpoint_port"],
                TransportLayerProtocol.UDP,
            ):
                self.logger.debug(
                    f"Creating new UDP endpoint for {message['client_endpoint_ip']}:{message['client_endpoint_port']}"
                )

                udp_endpoint = await self._endpoint_factory.create_server_endpoint(
                    Endpoint(
                        ip=ipaddress.IPv4Address(message["client_endpoint_ip"]),
                        port=message["client_endpoint_port"],
                    ),
                    TransportLayerProtocol.UDP,
                    self._someip_message_callback,
                )

                self._someip_client_endpoints.add_endpoint(client_id, udp_endpoint)

        if message["tcp"]:
            protocols.append(TransportLayerProtocol.TCP)

        event_group = EventGroup.from_json(message["eventgroup"])

        client_address = ipaddress.IPv4Address(message["client_endpoint_ip"])
        client_endpoint = Endpoint(
            ip=client_address, port=message["client_endpoint_port"]
        )

        new_subscription = Subscription(
            service_id=message["service_id"],
            instance_id=message["instance_id"],
            major_version=message["major_version"],
            client_endpoint=client_endpoint,
            server_endpoint=None,
            protocols=frozenset(protocols),
            eventgroup=event_group,
            ttl_seconds=message["ttl_subscription"],
        )

        if new_subscription in self._requested_subscriptions.subscriptions:
            self.logger.warning(
                f"The requested subscription received by client {client_id} is already requested."
            )

        self.logger.debug(f"Add subscription to storage with id {client_id}")
        self._requested_subscriptions.add_subscription(client_id, new_subscription)

    def _handle_stop_subscribe_eventgroup_request(
        self, message: StopSubscribeEventGroupRequest, client_id: int
    ):
        client_endpoint = Endpoint(
            ip=ipaddress.IPv4Address(message["client_endpoint_ip"]),
            port=message["client_endpoint_port"],
        )

        event_group = EventGroup.from_json(message["eventgroup"])

        protocols = [
            protocol
            for flag, protocol in (
                (message["udp"], TransportLayerProtocol.UDP),
                (message["tcp"], TransportLayerProtocol.TCP),
            )
            if flag
        ]

        subscription_to_remove = Subscription(
            service_id=message["service_id"],
            instance_id=message["instance_id"],
            major_version=message["major_version"],
            client_endpoint=client_endpoint,
            server_endpoint=None,
            protocols=frozenset(protocols),
            eventgroup=event_group,
            ttl_seconds=0,  # TTL is not relevant for removal
        )

        self._requested_subscriptions.remove_subscription(
            client_id, subscription_to_remove
        )

        if (
            len(self._requested_subscriptions.get_client_ids(subscription_to_remove))
            == 0
        ):

            for pending_subscription in list(self._pending_subscriptions):
                if (
                    pending_subscription.service_id == subscription_to_remove.service_id
                    and pending_subscription.instance_id
                    == subscription_to_remove.instance_id
                    and pending_subscription.major_version
                    == subscription_to_remove.major_version
                    and pending_subscription.eventgroup.id
                    == subscription_to_remove.eventgroup.id
                    and pending_subscription.client_endpoint
                    == subscription_to_remove.client_endpoint
                ):
                    self._pending_subscriptions.remove(pending_subscription)

            for active_subscription in list(self._active_subscriptions):
                if (
                    active_subscription.service_id == subscription_to_remove.service_id
                    and active_subscription.instance_id
                    == subscription_to_remove.instance_id
                    and active_subscription.major_version
                    == subscription_to_remove.major_version
                    and active_subscription.eventgroup.id
                    == subscription_to_remove.eventgroup.id
                    and active_subscription.client_endpoint
                    == subscription_to_remove.client_endpoint
                ):
                    self._active_subscriptions.remove(active_subscription)

                    (
                        session_id,
                        reboot_flag,
                    ) = self._unicast_session_handler.update_session()

                    # Build subscribe message
                    sd_message = SdMessage()
                    sd_message.session_id = session_id
                    sd_message.reboot_flag = reboot_flag

                    options = []
                    for protocol in active_subscription.protocols:
                        options.append(
                            IpV4EndpointOption(
                                address=active_subscription.client_endpoint.ip,
                                protocol=protocol,
                                port=active_subscription.client_endpoint.port,
                            )
                        )

                    entry = StopSubscribeEventGroupEntry(
                        service_id=active_subscription.service_id,
                        instance_id=active_subscription.instance_id,
                        major_version=active_subscription.major_version,
                        eventgroup_id=active_subscription.eventgroup.id,
                        counter=0,
                        ip_v4_endpoints=options,
                        ip_v6_endpoints=[],
                    )
                    sd_message.entries.append(entry)

                    # Send to server
                    if self._ucast_transport:
                        self._ucast_transport.sendto(
                            serialize_sd_message(sd_message),
                            (
                                str(active_subscription.server_endpoint.ip),
                                self.sd_port,
                            ),
                        )

    async def _handle_offer_service_request(
        self, message: OfferServiceRequest, client_id: int
    ):
        method_strs = message.get("method_list", [])
        methods = [Method.from_json(m) for m in method_strs]

        eventgroup_strs = message.get("eventgroup_list", [])
        eventgroups = [EventGroup.from_json(e) for e in eventgroup_strs]

        """
        - Store offered service including events
        - Subscribe received with service id, instance id, major version, ttl, eventgroup id. References an endpoint with ip and port option
        - If service is offered and found, send back an acknowledge and store the subscription with:
            - endpoint ip and port of the client where the events shall be sent to
            - service id, instance id, major version, eventgroup id,
        
        """

        service_to_add = ServiceToOffer(
            client_writer_id=client_id,
            instance_id=message["instance_id"],
            service_id=message["service_id"],
            major_version=message["major_version"],
            minor_version=message["minor_version"],
            offer_ttl_seconds=message["ttl"],
            cyclic_offer_delay_ms=message["cyclic_offer_delay_ms"],
            endpoint=Endpoint(
                ipaddress.IPv4Address(message["endpoint_ip"]), message["endpoint_port"]
            ),
            methods=methods,
            eventgroups=eventgroups,
        )

        self._services_to_offer.add_service(service_to_add)

        # Check if there is already an endpoint for the ip and port, if not, open a new endpoint
        if service_to_add.has_udp:
            if not self._someip_server_endpoints.has_endpoint(
                str(service_to_add.endpoint.ip),
                service_to_add.endpoint.port,
                TransportLayerProtocol.UDP,
            ):
                self.logger.debug(
                    f"Creating new UDP endpoint for {service_to_add.endpoint}"
                )

                udp_endpoint = await self._endpoint_factory.create_server_endpoint(
                    service_to_add.endpoint,
                    TransportLayerProtocol.UDP,
                    self._someip_message_callback,
                )

                self._someip_server_endpoints.add_endpoint(client_id, udp_endpoint)

        if service_to_add.has_tcp:
            if not self._someip_server_endpoints.has_endpoint(
                str(service_to_add.endpoint.ip),
                service_to_add.endpoint.port,
                TransportLayerProtocol.TCP,
            ):
                self.logger.debug(
                    f"Creating new TCP endpoint for {service_to_add.endpoint}"
                )

                tcp_endpoint = await self._endpoint_factory.create_server_endpoint(
                    service_to_add.endpoint,
                    TransportLayerProtocol.TCP,
                    self._someip_message_callback,
                )

                self._someip_server_endpoints.add_endpoint(client_id, tcp_endpoint)

        cyclic_offer_delay_ms = message["cyclic_offer_delay_ms"]

        # If there is no timer running for the interval yet, create a new timer task
        if cyclic_offer_delay_ms not in self._offer_timers:
            self.logger.debug(f"Starting new offer timer for {cyclic_offer_delay_ms}ms")
            self._offer_timers[cyclic_offer_delay_ms] = SimplePeriodicTimer(
                cyclic_offer_delay_ms / 1000.0,
                functools.partial(self.offer_timer_callback, cyclic_offer_delay_ms),
            )
            self._offer_timers[cyclic_offer_delay_ms].start()

    def _handle_stop_offer_service_request(
        self, message: StopOfferServiceRequest, writer_id: int
    ):
        method_strs = message.get("method_list", [])
        methods = [Method.from_json(m) for m in method_strs]

        eventgroup_strs = message.get("eventgroup_list", [])
        eventgroups = [EventGroup.from_json(e) for e in eventgroup_strs]

        service_to_stop = ServiceToOffer(
            client_writer_id=writer_id,
            instance_id=message["instance_id"],
            service_id=message["service_id"],
            major_version=message["major_version"],
            minor_version=message["minor_version"],
            offer_ttl_seconds=message["ttl"],
            cyclic_offer_delay_ms=message["cyclic_offer_delay_ms"],
            endpoint=Endpoint(
                ipaddress.IPv4Address(message["endpoint_ip"]), message["endpoint_port"]
            ),
            methods=methods,
            eventgroups=eventgroups,
        )

        # Remove the service from the storage
        self._services_to_offer.remove_service(service_to_stop)
        self._cleanup_unused_timers()

        (
            session_id,
            reboot_flag,
        ) = self._mcast_session_handler.update_session()

        sd_message = create_stop_offer_service_message(
            services_to_stop=[service_to_stop],
            session_id=session_id,
            reboot_flag=reboot_flag,
        )

        if self._ucast_transport:
            self.logger.debug(
                f"Send stop offer message for service 0x{service_to_stop.service_id:04x}, instance 0x{service_to_stop.instance_id:04x} to {self.sd_address}:{self.sd_port}"
            )
            self._ucast_transport.sendto(
                serialize_sd_message(sd_message), (self.sd_address, self.sd_port)
            )

        if service_to_stop.has_udp:
            try:
                udp_endpoint = self._someip_server_endpoints.get_endpoint(
                    writer_id, TransportLayerProtocol.UDP
                )
                if udp_endpoint:
                    self.logger.debug(
                        f"Closing UDP endpoint for {service_to_stop.endpoint}"
                    )
                    udp_endpoint.shutdown()
                    self._someip_server_endpoints.remove_endpoint(
                        writer_id, udp_endpoint
                    )
            except Exception as e:
                self.logger.error(
                    f"Error closing UDP endpoint for {service_to_stop.endpoint}: {e}"
                )

        if service_to_stop.has_tcp:
            try:
                tcp_endpoint = self._someip_server_endpoints.get_endpoint(
                    writer_id, TransportLayerProtocol.TCP
                )
                if tcp_endpoint:
                    self.logger.debug(
                        f"Closing TCP endpoint for {service_to_stop.endpoint}"
                    )
                    tcp_endpoint.shutdown()
                    self._someip_server_endpoints.remove_endpoint(
                        writer_id, tcp_endpoint
                    )
            except Exception as e:
                self.logger.error(
                    f"Error closing TCP endpoint for {service_to_stop.endpoint}: {e}"
                )

    def _handle_inbound_call_method_response(
        self, message: InboundCallMethodResponse, writer_id: int
    ):
        self.logger.debug(f"Received CallMethodResponse: {message}")

        header = SomeIpHeader(
            service_id=message["service_id"],
            method_id=message["method_id"],
            length=0,
            client_id=message["client_id"],
            session_id=message["session_id"],
            protocol_version=message["protocol_version"],
            interface_version=message["interface_version"],
            message_type=message["message_type"],
            return_code=message["return_code"],
        )

        payload_decoded = base64.b64decode(message["payload"])
        header.length = 8 + len(payload_decoded)

        endpoint = self._someip_server_endpoints.get_endpoint(
            writer_id, TransportLayerProtocol(message["protocol"])
        )
        self.logger.debug(
            f"Sending CallMethodResponse to {message['src_endpoint_ip']}:{message['src_endpoint_port']}"
        )

        if endpoint:
            endpoint.sendto(
                header.to_buffer() + payload_decoded,
                (message["src_endpoint_ip"], message["src_endpoint_port"]),
            )

    async def _handle_outbound_call_method_request(
        self, message: OutboundCallMethodRequest, client_id: int
    ):
        endpoint = None
        if TransportLayerProtocol(message["protocol"]) == TransportLayerProtocol.UDP:
            if not self._someip_client_endpoints.has_endpoint(
                message["src_endpoint_ip"],
                message["src_endpoint_port"],
                TransportLayerProtocol.UDP,
            ):
                self.logger.debug(
                    f"Creating new UDP endpoint for {message['src_endpoint_ip']}:{message['src_endpoint_port']}"
                )

                dst_endpoint = Endpoint(
                    ip=ipaddress.IPv4Address(message["dst_endpoint_ip"]),
                    port=message["dst_endpoint_port"],
                )
                src_endpoint = Endpoint(
                    ip=ipaddress.IPv4Address(message["src_endpoint_ip"]),
                    port=message["src_endpoint_port"],
                )

                udp_endpoint = await self._endpoint_factory.create_udp_client_endpoint(
                    dst_endpoint,
                    src_endpoint,
                    self._someip_message_callback,
                    self.logger,
                )
                self._someip_client_endpoints.add_endpoint(client_id, udp_endpoint)

                endpoint = udp_endpoint
            else:
                endpoint = self._someip_client_endpoints.get_endpoint_by_ip_port(
                    message["src_endpoint_ip"],
                    message["src_endpoint_port"],
                    TransportLayerProtocol.UDP,
                )

        elif TransportLayerProtocol(message["protocol"]) == TransportLayerProtocol.TCP:

            if not self._someip_client_endpoints.has_tcp_endpoint(
                message["src_endpoint_ip"],
                message["src_endpoint_port"],
                message["dst_endpoint_ip"],
                message["dst_endpoint_port"],
            ):
                self.logger.debug(
                    f"Creating new TCP endpoint for {message['src_endpoint_ip']}:{message['src_endpoint_port']}"
                )

                tcp_endpoint = self._endpoint_factory.create_tcp_client_endpoint(
                    Endpoint(
                        ip=ipaddress.IPv4Address(message["dst_endpoint_ip"]),
                        port=message["dst_endpoint_port"],
                    ),
                    Endpoint(
                        ip=ipaddress.IPv4Address(message["src_endpoint_ip"]),
                        port=message["src_endpoint_port"],
                    ),
                    self._someip_message_callback,
                    self.logger,
                )

                self._someip_client_endpoints.add_endpoint(client_id, tcp_endpoint)
                endpoint: TCPClientSomeipEndpoint = tcp_endpoint
            else:
                endpoint: TCPClientSomeipEndpoint = (
                    self._someip_client_endpoints.get_end_point_by_src_and_dst_ip_port(
                        message["src_endpoint_ip"],
                        message["src_endpoint_port"],
                        message["dst_endpoint_ip"],
                        message["dst_endpoint_port"],
                        TransportLayerProtocol.TCP,
                    )
                )

            # TODO: This shall not block the handle_client function. A new task shall be created
            # For TCP wait for the connection to be established
            while not endpoint.is_connected():
                # self.logger.debug(
                #    f"Waiting for TCP connection to {message['dst_endpoint_ip']}:{message['dst_endpoint_port']}"
                # )
                await asyncio.sleep(0.2)

        # Build the request message
        self.logger.debug(
            f"Sending OutboundCallMethodRequest to {message['dst_endpoint_ip']}:{message['dst_endpoint_port']}"
        )

        decoded_payload = base64.b64decode(message["payload"])

        header = SomeIpHeader(
            service_id=message["service_id"],
            method_id=message["method_id"],
            client_id=message["client_id"],
            session_id=message["session_id"],
            protocol_version=0x01,
            interface_version=message["major_version"],
            message_type=MessageType.REQUEST.value,
            return_code=0x00,
            length=len(decoded_payload) + 8,
        )
        someip_message = SomeIpMessage(header, decoded_payload)

        new_call = MethodCall(
            service_id=message["service_id"],
            method_id=message["method_id"],
            client_id=message["client_id"],
            session_id=message["session_id"],
            src_ip=message["src_endpoint_ip"],
            src_port=message["src_endpoint_port"],
        )

        if new_call in self._issued_method_calls:
            self.logger.warning(
                f"Method call {new_call} already issued. Overwriting writer_id."
            )

        self._issued_method_calls[new_call] = client_id

        endpoint.sendto(
            someip_message.serialize(),
            (message["dst_endpoint_ip"], message["dst_endpoint_port"]),
        )

    def _handle_find_service_request(self, message: FindServiceRequest, writer_id: int):

        service_found = False
        all_services = [s for s in self._found_services]

        for service in self._services_to_offer.get_all_services():
            protocols_to_add = set()
            if service.has_udp:
                protocols_to_add.add(TransportLayerProtocol.UDP)
            if service.has_tcp:
                protocols_to_add.add(TransportLayerProtocol.TCP)

            service_to_add = ServiceInstance(
                service_id=service.service_id,
                instance_id=service.instance_id,
                major_version=service.major_version,
                minor_version=service.minor_version,
                ttl=service.offer_ttl_seconds,
                endpoint=service.endpoint,
                protocols=frozenset(protocols_to_add),
                timestamp=0.0,
            )

            all_services.append(service_to_add)

        for found_service in all_services:
            """
            • Instance ID shall be set to 0xFFFF, if all service instances shall be returned. It
            shall be set to the Instance ID of a specific service instance, if just a single service
            instance shall be returned.
            • Major Version shall be set to 0xFF, that means that services with any version shall
            be returned. If set to value different than 0xFF, services with this specific major
            version shall be returned only.
            • Minor Version shall be set to 0xFFFF FFFF, that means that services with any
            version shall be returned. If set to a value different to 0xFFFF FFFF, services
            with this specific minor version shall be returned only
            """

            if (
                (message["service_id"] == found_service.service_id)
                and (
                    message["instance_id"] == found_service.instance_id
                    or message["instance_id"] == 0xFFFF
                )
                and (
                    message["major_version"] == found_service.major_version
                    or message["major_version"] == 0xFF
                )
                and (
                    message["minor_version"] == found_service.minor_version
                    or message["minor_version"] == 0xFFFFFFFF
                )
            ):
                service_found = True
                response = create_uds_message(
                    FindServiceResponse,
                    success=True,
                    service_id=found_service.service_id,
                    instance_id=found_service.instance_id,
                    major_version=found_service.major_version,
                    minor_version=found_service.minor_version,
                    endpoint_ip=str(found_service.endpoint.ip),
                    endpoint_port=found_service.endpoint.port,
                )

                tx_queue = self._tx_queues.get(writer_id)
                if tx_queue:
                    tx_queue.put_nowait(self.prepare_message(response))

                break

        if not service_found:
            response = create_uds_message(
                FindServiceResponse,
                success=False,
                service_id=message["service_id"],
                instance_id=message["instance_id"],
                major_version=message["major_version"],
                minor_version=message["minor_version"],
                endpoint_ip="empty",
                endpoint_port=0,
            )

            tx_queue = self._tx_queues.get(writer_id)
            if tx_queue:
                tx_queue.put_nowait(self.prepare_message(response))

    def _handle_send_event_request(self, message: SendEventRequest, writer_id: int):
        for sub in self._service_subscribers.values():
            sub.update()

        deserialized_event = Event.from_json(message["event"])
        payload_decoded = base64.b64decode(message["payload"])

        for offered_service in self._service_subscribers.keys():
            if (
                offered_service.service_id == message["service_id"]
                and offered_service.instance_id == message["instance_id"]
                and message["eventgroup_id"] in offered_service.eventgroup_ids
            ):

                if self._service_subscribers[offered_service].has_subscribers:
                    for subscriber in self._service_subscribers[
                        offered_service
                    ].subscribers:
                        self.logger.debug(
                            f"Sending event to subscriber {subscriber.endpoint[0]}:{subscriber.endpoint[1]}"
                        )

                        if deserialized_event.protocol == TransportLayerProtocol.UDP:
                            endpoint = (
                                self._someip_server_endpoints.get_endpoint_by_ip_port(
                                    message["src_endpoint_ip"],
                                    message["src_endpoint_port"],
                                    TransportLayerProtocol.UDP,
                                )
                            )

                            if endpoint:
                                someip_header = SomeIpHeader(
                                    service_id=offered_service.service_id,
                                    method_id=deserialized_event.id,
                                    length=len(payload_decoded) + 8,
                                    client_id=message["client_id"],
                                    session_id=message["session_id"],
                                    protocol_version=1,
                                    interface_version=offered_service.major_version,
                                    message_type=MessageType.NOTIFICATION.value,
                                    return_code=0x00,
                                )

                                someip_message = SomeIpMessage(
                                    someip_header, payload_decoded
                                )
                                endpoint.sendto(
                                    someip_message.serialize(), subscriber.endpoint
                                )

                        elif deserialized_event.protocol == TransportLayerProtocol.TCP:

                            endpoint: TCPSomeipEndpoint = (
                                self._someip_server_endpoints.get_endpoint_by_ip_port(
                                    message["src_endpoint_ip"],
                                    message["src_endpoint_port"],
                                    TransportLayerProtocol.TCP,
                                )
                            )

                            if endpoint:
                                someip_header = SomeIpHeader(
                                    service_id=offered_service.service_id,
                                    method_id=deserialized_event.id,
                                    length=len(payload_decoded) + 8,
                                    client_id=message["client_id"],
                                    session_id=message["session_id"],
                                    protocol_version=1,
                                    interface_version=offered_service.major_version,
                                    message_type=MessageType.NOTIFICATION.value,
                                    return_code=0x00,
                                )

                                someip_message = SomeIpMessage(
                                    someip_header, payload_decoded
                                )
                                endpoint.sendto(
                                    someip_message.serialize(), subscriber.endpoint
                                )

        # Handle internal subscriptions (UDS clients)
        for requested_subscription in self._requested_subscriptions.subscriptions:
            if (
                requested_subscription.service_id == message["service_id"]
                and requested_subscription.instance_id == message["instance_id"]
                and requested_subscription.eventgroup.id == message["eventgroup_id"]
                and requested_subscription.major_version == message["major_version"]
            ):
                self.logger.debug(
                    f"Internal subscription for service {message['service_id']:04x}, "
                    f"instance {message['instance_id']:04x}, eventgroup {message['eventgroup_id']:04x}"
                )

                event_msg = create_uds_message(
                    ReceivedEvent,
                    service_id=message["service_id"],
                    event_id=deserialized_event.id,
                    src_endpoint_ip=message["src_endpoint_ip"],
                    src_endpoint_port=message["src_endpoint_port"],
                    payload=message["payload"],
                )

                writer_ids = self._requested_subscriptions.get_client_ids(
                    requested_subscription
                )

                for writer_id in writer_ids:
                    tx_queue = self._tx_queues.get(writer_id)
                    if tx_queue:
                        tx_queue.put_nowait(self.prepare_message(event_msg))

    def offer_timer_callback(self, cyclic_offer_delay_ms: int):
        self.logger.debug(f"Offer timer callback for {cyclic_offer_delay_ms}ms")

        services_to_offer: List[ServiceToOffer] = (
            self._services_to_offer.services_by_cyclic_offer_delay(
                cyclic_offer_delay_ms
            )
        )

        for service in services_to_offer:
            service.last_offer_time = time.time()

        if len(services_to_offer) > 0:
            (
                session_id,
                reboot_flag,
            ) = self._mcast_session_handler.update_session()

            sd_message = create_offer_service_message(
                services_to_offer=services_to_offer,
                session_id=session_id,
                reboot_flag=reboot_flag,
            )

            if self._ucast_transport:
                self._ucast_transport.sendto(
                    serialize_sd_message(sd_message), (self.sd_address, self.sd_port)
                )

    def prepare_message(self, message: dict):
        payload = json.dumps(message).encode("utf-8")
        return struct.pack("<I", len(payload)) + bytes(256 - 4) + payload

    async def send_to_all_clients(self, message):
        for writer in self.clients:
            try:
                writer.write(message)
                await writer.drain()
            except ConnectionError:
                self.logger.error("Error sending message to client.")
                self.clients.remove(writer)
                writer.close()
                await writer.wait_closed()

    async def start_server(self):
        try:
            if self._ttl_task is None or self._ttl_task.done():
                self._ttl_task = asyncio.create_task(self._check_services_ttl_task())
            await self.start_sd_listening()
            await self._server.serve_forever()
        except asyncio.CancelledError:
            self.logger.info(f"Server cancelled.")
        finally:
            if self._mcast_transport:
                self._mcast_transport.close()
            if self._ucast_transport:
                self._ucast_transport.close()

            self.logger.info(f"Server stopped.")

    def _timeout_of_offered_service(self, offered_service: SdService):
        self.logger.info(
            f"Offered service timed out: service id 0x{offered_service.service_id:04x}, instance id 0x{offered_service.instance_id:04x}"
        )

        # TODO: If clients subscribed to this service, remove all subscriptions

    async def wait_for_message_in_rx_queue(
        self, rx_queue: asyncio.Queue, msg_type: str, timeout: int = 1.0
    ) -> dict:

        messages_to_keep = []
        found_message = None
        message_found = False

        while not message_found:
            try:
                msg = await asyncio.wait_for(rx_queue.get(), timeout=timeout)

                if msg.get("type") == msg_type:
                    found_message = msg
                    message_found = True
                else:
                    messages_to_keep.append(msg)
            except asyncio.TimeoutError:
                # Put all non-target messages back in the queue
                for msg in messages_to_keep:
                    await rx_queue.put(msg)
                return None

        # Put all non-target messages back in the queue
        for msg in messages_to_keep:
            await rx_queue.put(msg)

        return found_message

    def _handle_offered_service(self, offered_service: ServiceInstance):
        self.logger.info(f"Received offered service: {offered_service}")

        if offered_service not in self._found_services:
            self._found_services.append(offered_service)
        else:
            # Update the timestamp if the service is already in the list
            index = self._found_services.index(offered_service)
            self._found_services[index].timestamp = offered_service.timestamp

        # Check if there is a requested subscription for this service
        for (
            requested_subscription,
            client_id,
        ) in self._requested_subscriptions.has_subscriptions(
            offered_service.service_id,
            offered_service.instance_id,
            offered_service.major_version,
        ):

            requested_protocols: Set[TransportLayerProtocol] = (
                offered_service.protocols & requested_subscription.protocols
            )

            if TransportLayerProtocol.TCP in requested_protocols:
                if not self._someip_client_endpoints.has_tcp_endpoint(
                    str(requested_subscription.client_endpoint.ip),
                    requested_subscription.client_endpoint.port,
                    str(offered_service.endpoint.ip),
                    offered_service.endpoint.port,
                ):
                    self.logger.debug(
                        f"Creating new TCP endpoint for {requested_subscription.client_endpoint}"
                    )

                    tcp_endpoint = self._endpoint_factory.create_tcp_client_endpoint(
                        offered_service.endpoint,
                        requested_subscription.client_endpoint,
                        self._someip_message_callback,
                        self.logger,
                    )

                    self._someip_client_endpoints.add_endpoint(client_id, tcp_endpoint)

                    # TODO: This shall not block the handle_client function. A new task shall be created
                    # For TCP wait for the connection to be established
                    # while not tcp_endpoint.is_connected():
                    #    await asyncio.sleep(0.2)

            (
                session_id,
                reboot_flag,
            ) = self._unicast_session_handler.update_session()

            # Build subscribe message
            sd_message = SdMessage()
            sd_message.session_id = session_id
            sd_message.reboot_flag = reboot_flag

            options = []
            for protocol in requested_protocols:
                options.append(
                    IpV4EndpointOption(
                        address=requested_subscription.client_endpoint.ip,
                        protocol=protocol,
                        port=requested_subscription.client_endpoint.port,
                    )
                )

            entry = SubscribeEventGroupEntry(
                service_id=offered_service.service_id,
                instance_id=offered_service.instance_id,
                major_version=offered_service.major_version,
                ttl=requested_subscription.ttl_seconds,
                eventgroup_id=requested_subscription.eventgroup.id,
                counter=0,
                ip_v4_endpoints=options,
                ip_v6_endpoints=[],
            )
            sd_message.entries.append(entry)

            client_endpoint = requested_subscription.client_endpoint
            server_endpoint = offered_service.endpoint

            pending_subscription = Subscription(
                service_id=offered_service.service_id,
                instance_id=offered_service.instance_id,
                major_version=offered_service.major_version,
                eventgroup=requested_subscription.eventgroup,
                ttl_seconds=requested_subscription.ttl_seconds,
                client_endpoint=client_endpoint,
                server_endpoint=server_endpoint,
                protocols=frozenset(requested_protocols),
                timestamp_last_update=time.time(),
            )
            self._pending_subscriptions.add(pending_subscription)

            if self._ucast_transport:
                self._ucast_transport.sendto(
                    serialize_sd_message(sd_message),
                    (str(offered_service.endpoint.ip), self.sd_port),
                )

    def _handle_subscription(
        self,
        sd_subscription: SdSubscription,
    ):
        # TODO: Send back a nack message if no service is found
        self.logger.info(f"Received subscription: {sd_subscription}")

        for sub in self._service_subscribers.values():
            sub.update()

        # From SD specification:
        # [PRS_SOMEIPSD_00829] When receiving a SubscribeEventgroupAck or Sub-
        # scribeEventgroupNack the Service ID, Instance ID, Eventgroup ID, and Major Ver-
        # sion shall match exactly to the corresponding SubscribeEventgroup Entry to identify
        # an Eventgroup of a Service Instance.
        # Check if the service id, the instance and the major version is in the list of offered services
        # If yes, check if the eventgroup id is in the list of eventgroup ids
        # If yes, subscribe to the eventgroup
        for offered_service in self._services_to_offer.get_all_services():
            # [PRS_SOMEIPSD_00828] When receiving a SubscribeEventgroup or StopSubscribeEventgroup the Service ID,
            # Instance ID, Eventgroup ID, and Major Version shall
            # match exactly to the configured values to identify an Eventgroup of a Service Instance.
            if (
                offered_service.service_id == sd_subscription.service_id
                and offered_service.instance_id == sd_subscription.instance_id
                and sd_subscription.eventgroup_id in offered_service.eventgroup_ids
                and offered_service.major_version == sd_subscription.major_version
            ):

                self.logger.info(
                    f"Subscription to eventgroup 0x{sd_subscription.eventgroup_id:04X} of service 0x{offered_service.service_id:04X}, instance 0x{offered_service.instance_id:04X} requested."
                )

                (
                    session_id,
                    reboot_flag,
                ) = self._unicast_session_handler.update_session()

                ack_entry = build_subscribe_eventgroup_ack_entry(
                    service_id=offered_service.service_id,
                    instance_id=offered_service.instance_id,
                    major_version=offered_service.major_version,
                    ttl=sd_subscription.ttl,
                    event_group_id=sd_subscription.eventgroup_id,
                )

                header_output = build_subscribe_eventgroup_ack_sd_header(
                    entry=ack_entry,
                    session_id=session_id,
                    reboot_flag=reboot_flag,
                )

                self.logger.info(
                    f"Sending subscribe ack for eventgroup 0x{sd_subscription.eventgroup_id:04X} of service 0x{offered_service.service_id:04X} instance 0x{offered_service.instance_id:04X} to {sd_subscription.ipv4_address}:{sd_subscription.port}"
                )

                new_subscriber = EventGroupSubscriber(
                    sd_subscription.eventgroup_id,
                    (sd_subscription.ipv4_address, sd_subscription.port),
                    sd_subscription.ttl,
                )

                if offered_service not in self._service_subscribers:
                    self._service_subscribers[offered_service] = Subscribers()

                self._service_subscribers[offered_service].add_subscriber(
                    new_subscriber
                )

                if self._ucast_transport:
                    self._ucast_transport.sendto(
                        data=header_output.to_buffer(),
                        addr=(
                            str(sd_subscription.ipv4_address),
                            self.sd_port,
                        ),
                    )

    def _handle_sd_subscribe_ack_eventgroup_entry(
        self, event_group_entry: SdEventGroupEntry
    ):
        self.logger.info(
            f"Received subscribe ack eventgroup entry: {event_group_entry}"
        )
        pending_subscription = None

        for pending_subscription_tmp in self._pending_subscriptions:
            self.logger.debug(
                f"Checking pending subscription: {pending_subscription_tmp} with values \
                service_id={pending_subscription_tmp.service_id}, \
                instance_id={pending_subscription_tmp.instance_id}, \
                major_version={pending_subscription_tmp.major_version}, \
                eventgroup_id={pending_subscription_tmp.eventgroup.id}"
            )

            self.logger.debug(
                f"Event group entry values: \
                service_id={event_group_entry.sd_entry.service_id}, \
                instance_id={event_group_entry.sd_entry.instance_id}, \
                major_version={event_group_entry.sd_entry.major_version}, \
                eventgroup_id={event_group_entry.eventgroup_id}"
            )

            if (
                pending_subscription_tmp.service_id
                == event_group_entry.sd_entry.service_id
                and pending_subscription_tmp.instance_id
                == event_group_entry.sd_entry.instance_id
                and pending_subscription_tmp.major_version
                == event_group_entry.sd_entry.major_version
                and pending_subscription_tmp.eventgroup.id
                == event_group_entry.eventgroup_id
            ):
                self.logger.debug(
                    f"Found matching pending subscription: {pending_subscription_tmp}"
                )
                pending_subscription = pending_subscription_tmp
                break

        if pending_subscription is not None:
            pending_subscription.timestamp_last_update = time.time()
            self._active_subscriptions.discard(pending_subscription)
            self._active_subscriptions.add(pending_subscription)

            self.logger.info(f"Subscription acknowledged: {pending_subscription}")
            self._pending_subscriptions.discard(pending_subscription)

    def _handle_sd_subscribe_nack_eventgroup_entry(
        self, event_group_entry: SdEventGroupEntry
    ):
        self.logger.info(
            f"Received subscribe nack eventgroup entry: {event_group_entry}"
        )

    def datagram_received_mcast(
        self, data: bytes, addr: Tuple[Union[str, Any], int]
    ) -> None:

        # Ignore messages from the same interface and port
        if addr[0] == self.interface and addr[1] == self.sd_port:
            return

        if not is_sd_message(data):
            return

        sd_message = deserialize_sd_message(data, addr[0], addr[1], multicast=True)
        sd_message.timestamp = time.time()

        for offer_service_entry in [
            o
            for o in sd_message.entries
            if o.type
            == someipy._internal._sd.entries.sd_entry.SdEntryType.OFFER_SERVICE
        ]:
            entry: OfferServiceEntry = offer_service_entry

            protocols = set()
            for ep in entry.ip_v4_endpoints:
                protocols.add(ep.protocol)
            for ep in entry.ip_v6_endpoints:
                protocols.add(ep.protocol)

            endpoint = Endpoint(
                ip=entry.ip_v4_endpoints[0].address,
                port=entry.ip_v4_endpoints[0].port,
            )

            service_instance = ServiceInstance(
                service_id=entry.service_id,
                instance_id=entry.instance_id,
                major_version=entry.major_version,
                minor_version=entry.minor_version,
                ttl=entry.ttl,
                endpoint=endpoint,
                protocols=frozenset(protocols),
                timestamp=sd_message.timestamp,
            )
            self._handle_offered_service(service_instance)

        someip_sd_header = SomeIpSdHeader.from_buffer(data)

        for subscription in extract_subscribe_entries(someip_sd_header):
            self._handle_subscription(subscription)

        for event_group_entry in extract_subscribe_ack_eventgroup_entries(
            someip_sd_header
        ):
            self._handle_sd_subscribe_ack_eventgroup_entry(event_group_entry)

    def connection_lost_mcast(self, exc: Exception) -> None:
        pass

    def datagram_received_ucast(
        self, data: bytes, addr: Tuple[Union[str, Any], int]
    ) -> None:
        self.logger.debug(f"Received SD message from {addr}: {data}")

        # Ignore messages from the same interface and port
        if addr[0] == self.interface and addr[1] == self.sd_port:
            return

        someip_header = SomeIpHeader.from_buffer(data)
        if not someip_header.is_sd_header():
            return

        someip_sd_header = SomeIpSdHeader.from_buffer(data)

        for subscription in extract_subscribe_entries(someip_sd_header):
            self._handle_subscription(subscription)

        for event_group_entry in extract_subscribe_ack_eventgroup_entries(
            someip_sd_header
        ):
            self._handle_sd_subscribe_ack_eventgroup_entry(event_group_entry)

        for event_group_entry in extract_subscribe_nack_eventgroup_entries(
            someip_sd_header
        ):
            self._handle_sd_subscribe_nack_eventgroup_entry(event_group_entry)

    def connection_lost_ucast(self, exc: Exception) -> None:
        pass

    async def start_sd_listening(self):
        if self.sd_address.startswith("224"):
            self._sd_socket_mcast = create_rcv_multicast_socket(
                self.sd_address, self.sd_port, self.interface
            )
        else:
            self._sd_socket_mcast = create_rcv_broadcast_socket(
                self.sd_address, self.sd_port, self.interface
            )

        loop = asyncio.get_running_loop()
        self._mcast_transport, _ = await loop.create_datagram_endpoint(
            lambda: DatagramAdapter(
                target=None,
                datagram_received_callback=self.datagram_received_mcast,
                connection_lost_callback=self.connection_lost_mcast,
            ),
            sock=self._sd_socket_mcast,
        )

        self._sd_socket_ucast = create_udp_socket(self.interface, self.sd_port)
        self._ucast_transport, _ = await loop.create_datagram_endpoint(
            lambda: DatagramAdapter(
                target=None,
                datagram_received_callback=self.datagram_received_ucast,
                connection_lost_callback=self.connection_lost_ucast,
            ),
            sock=self._sd_socket_ucast,
        )


def _configure_logging(log_level=logging.DEBUG, log_path=None) -> logging.Logger:
    logger = logging.getLogger(f"someipyd")
    logger.setLevel(log_level)

    # Remove any existing handlers to prevent duplicate logs
    if logger.hasHandlers():
        logger.handlers.clear()

    formatter = logging.Formatter(
        "%(asctime)s.%(msecs)03d %(name)s [%(levelname)s]: %(message)s",
        datefmt="%Y-%m-%d,%H:%M:%S",
    )

    if log_path:
        file_handler = logging.FileHandler(log_path)
        file_handler.setLevel(log_level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    else:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setLevel(log_level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    return logger


def _load_config(config_file: str) -> dict:
    if config_file and os.path.exists(config_file):
        try:
            with open(config_file, "r") as f:
                return json.load(f)
        except (FileNotFoundError, json.JSONDecodeError) as e:
            print(f"Error loading config file: {e}. Using defaults.")
            return {}
    elif os.path.exists(DEFAULT_CONFIG_FILE):
        try:
            with open(DEFAULT_CONFIG_FILE, "r") as f:
                return json.load(f)
        except (FileNotFoundError, json.JSONDecodeError) as e:
            print(f"Error loading config file: {e}. Using defaults.")
            return {}
    else:
        return {}


async def async_main():
    parser = argparse.ArgumentParser(description="SOME/IP Daemon")
    parser.add_argument("--config", help="Path to configuration file")
    parser.add_argument("--log-path", help="Path to log file")
    args = parser.parse_args()

    # Load configuration
    config = _load_config(args.config)

    # Logging
    log_path = args.log_path if args.log_path else config.get("log_path", None)
    log_level = config.get("log_level", "INFO")

    log_level_mapping = {
        "DEBUG": logging.DEBUG,
        "ERROR": logging.ERROR,
        "INFO": logging.INFO,
        "FATAL": logging.FATAL,
    }

    if log_level in log_level_mapping:
        log_level = log_level_mapping[log_level]

    logger = _configure_logging(log_level=log_level, log_path=log_path)

    logger.info(
        f"Starting SOME/IP daemon with config:\n"
        f"Socket path: {config.get('socket_path', DEFAULT_SOCKET_PATH)}\n"
        f"SD address: {config.get('sd_address', DEFAULT_SD_ADDRESS)}\n"
        f"SD port: {config.get('sd_port', DEFAULT_SD_PORT)}\n"
        f"Interface: {config.get('interface', DEFAULT_INTERFACE_IP)}\n"
        f"Loglevel: {log_level}\n"
        f"Log path: {log_path if log_path else 'Console'}\n"
        f"Use tcp: {config.get('use_tcp', False)}\n"
        f"Tcp port: {config.get('tcp_port', None)}\n"
        f"Tcp host: {config.get('tcp_host', '127.0.0.1')}\n"
    )

    daemon_server = DaemonServer(logger)

    daemon = SomeipDaemon(daemon_server, SomeipEndpointFactory(), config, logger)

    daemon_server.client_connected += daemon.new_client_connected
    daemon_server.client_disconnected += daemon.client_disconnected

    await daemon_server.start(
        use_tcp=config.get("use_tcp", False),
        socket_path=config.get("socket_path", DEFAULT_SOCKET_PATH),
        tcp_port=config.get("tcp_port", 30500),
        host=config.get("tcp_host", "127.0.0.1"),
    )

    await daemon.start_server()


def main():
    try:
        asyncio.run(async_main())
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import time


class SimplePeriodicTimer:
    def __init__(self, period, callback):
        self._period = period
        self._callback = callback
        self._task = None
        self._running = False

    async def _job(self):
        try:
            while True:
                start = time.time()
                self._callback()
                elapsed = time.time() - start
                to_sleep = self._period - elapsed
                if elapsed > self._period:
                    to_sleep = 0
                await asyncio.sleep(to_sleep)
        except asyncio.CancelledError:
            self._running = False

    def start(self):
        if not self._running:
            self._running = True
            loop = asyncio.get_running_loop()
            self._task = loop.create_task(self._job())

    def stop(self):
        if self._running and self._task is not None:
            self._task.cancel()
            self._running = False

    @property
    def task(self):
        return self._task



import pytest
from someipy._internal.someip_data_processor import SomeipDataProcessor
from someipy._internal.someip_header import SomeIpHeader
from someipy._internal.message_types import MessageType
from someipy._internal.someip_message import SomeIpMessage


@pytest.fixture
def valid_someip_message() -> SomeIpMessage:
    # length is 8 bytes + payload size
    payload_size = 64
    length = 8 + payload_size

    someip_header = SomeIpHeader(
        service_id=1,
        method_id=2,
        length=length,
        client_id=3,
        session_id=4,
        protocol_version=1,
        interface_version=2,
        message_type=MessageType.REQUEST.value,
        return_code=0x00,
    )
    payload = b"\x00" * payload_size
    return SomeIpMessage(header=someip_header, payload=payload)


@pytest.fixture
def corrupt_someip_message() -> SomeIpMessage:
    # length is 8 bytes + payload size
    payload_size = 64
    length = 8 + payload_size + 1  # +1 for corrupt header

    someip_header = SomeIpHeader(
        service_id=1,
        method_id=2,
        length=length,
        client_id=3,
        session_id=4,
        protocol_version=1,
        interface_version=2,
        message_type=MessageType.REQUEST.value,
        return_code=0x00,
    )
    payload = b"\x00" * payload_size
    return SomeIpMessage(header=someip_header, payload=payload)


def test_process_with_datagrams(valid_someip_message):
    data = valid_someip_message.header.to_buffer() + valid_someip_message.payload
    processor = SomeipDataProcessor()
    result = processor.process_data(data)

    assert result is True
    assert processor.someip_message.header == valid_someip_message.header
    assert processor.expected_bytes == 8
    assert len(processor._buffer) == 0

    result = processor.process_data(data)
    assert result is True
    assert processor.someip_message.header == valid_someip_message.header
    assert processor._expected_bytes == 8
    assert len(processor._buffer) == 0


def test_process_with_malformed_datagrams(corrupt_someip_message):
    data = corrupt_someip_message.header.to_buffer() + corrupt_someip_message.payload
    processor = SomeipDataProcessor()

    result = processor.process_data(data)

    assert result is False



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, List, Optional
from someipy._internal.someip_endpoint import SomeipEndpoint
from someipy._internal.transport_layer_protocol import TransportLayerProtocol


class SomeipEndpointStorage:
    def __init__(self):
        self._storage: Dict[int, List[SomeipEndpoint]] = {}

    def add_endpoint(self, client_id: int, endpoint: SomeipEndpoint) -> bool:
        """
        Add an endpoint for a client ID.
        Returns True if added successfully, False if client already has 2 endpoints.
        """
        if client_id not in self._storage:
            self._storage[client_id] = [endpoint]
            return True

        if len(self._storage[client_id]) >= 2:
            return False

        self._storage[client_id].append(endpoint)
        return True

    def remove_endpoint(self, client_id: int, endpoint: SomeipEndpoint) -> bool:
        if client_id not in self._storage:
            return False

        try:
            self._storage[client_id].remove(endpoint)
            # Remove client entry if no endpoints left
            if not self._storage[client_id]:
                del self._storage[client_id]
            return True
        except ValueError:
            return False

    def remove_client(self, client_id: int) -> bool:
        if client_id in self._storage:
            del self._storage[client_id]
            return True
        return False

    def get_endpoints(self, client_id: int) -> Optional[List[SomeipEndpoint]]:
        return self._storage.get(client_id, None)

    def get_endpoint(
        self, client_id: int, protocol: TransportLayerProtocol
    ) -> Optional[SomeipEndpoint]:
        endpoints = self._storage.get(client_id, None)
        if endpoints is None:
            return None

        if any(endpoint.protocol() == protocol for endpoint in endpoints):
            # If the client has an endpoint with the same protocol, return all endpoints
            return [
                endpoint for endpoint in endpoints if endpoint.protocol() == protocol
            ][0]
        else:
            return None

    def get_endpoint_by_ip_port(
        self, src_ip: str, src_port: int, protocol: TransportLayerProtocol
    ) -> Optional[SomeipEndpoint]:
        """
        Get an endpoint by source IP, source port, and protocol.
        Returns the first matching endpoint or None if not found.
        """
        for endpoints in self._storage.values():
            for endpoint in endpoints:
                if (
                    endpoint.src_ip() == src_ip
                    and endpoint.src_port() == src_port
                    and endpoint.protocol() == protocol
                ):
                    return endpoint
        return None

    def get_end_point_by_src_and_dst_ip_port(
        self,
        src_ip: str,
        src_port: int,
        dst_ip: str,
        dst_port: int,
        protocol: TransportLayerProtocol,
    ) -> Optional[SomeipEndpoint]:
        """
        Get an endpoint by source IP, source port, destination IP, and destination port.
        Returns the first matching endpoint or None if not found.
        """
        for endpoints in self._storage.values():
            for endpoint in endpoints:
                if (
                    endpoint.src_ip() == src_ip
                    and endpoint.src_port() == src_port
                    and endpoint.dst_ip() == dst_ip
                    and endpoint.dst_port() == dst_port
                    and endpoint.protocol() == protocol
                ):
                    return endpoint
        return None

    def get_all_endpoints(self) -> List[SomeipEndpoint]:
        """Get all endpoints in the storage"""
        return [
            endpoint for endpoints in self._storage.values() for endpoint in endpoints
        ]

    def has_endpoint(
        self, src_ip: str, src_port: int, protocol: TransportLayerProtocol
    ) -> bool:
        """
        Check if an endpoint with the given IP, port, and protocol exists in the storage.
        """
        for endpoints in self._storage.values():
            for endpoint in endpoints:
                if (
                    endpoint.src_ip() == src_ip
                    and endpoint.src_port() == src_port
                    and endpoint.protocol() == protocol
                ):
                    return True
        return False

    def has_tcp_endpoint(
        self, src_ip: str, src_port: int, dst_ip: str, dst_port: int
    ) -> bool:
        """
        Check if a TCP endpoint with the given source IP, source port, destination IP, and destination port exists in the storage.
        """
        for endpoints in self._storage.values():
            for endpoint in endpoints:
                if (
                    endpoint.protocol() == TransportLayerProtocol.TCP
                    and endpoint.src_ip() == src_ip
                    and endpoint.src_port() == src_port
                    and endpoint.dst_ip() == dst_ip
                    and endpoint.dst_port() == dst_port
                ):
                    return True
        return False

    def __contains__(self, client_id: int) -> bool:
        """Check if a client ID exists in the storage"""
        return client_id in self._storage

    def __len__(self) -> int:
        """Get the number of clients in the storage"""
        return len(self._storage)



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from dataclasses import dataclass
from enum import Enum
import ipaddress
import socket
import struct


from someipy._internal._sd.entries.find_service_entry import FindServiceEntry
from someipy._internal._sd.entries.offer_service_entry import OfferServiceEntry
from someipy._internal._sd.entries.stop_offer_service_entry import StopOfferServiceEntry
from someipy._internal._sd.entries.stop_subscribe_eventgroup_entry import (
    StopSubscribeEventGroupEntry,
)
from someipy._internal._sd.entries.subscribe_ack_entry import (
    SubscribeAckEventGroupEntry,
)
from someipy._internal._sd.entries.subscribe_eventgroup_entry import (
    SubscribeEventGroupEntry,
)
from someipy._internal._sd.entries.subscribe_eventgroup_nack import (
    SubscribeEventGroupNackEntry,
)
from someipy._internal._sd.options.configuration_option import ConfigurationOption
from someipy._internal._sd.options.endpoint import (
    IpV4EndpointOption,
    IpV6EndpointOption,
)
from someipy._internal._sd.options.load_balancing import LoadBalancingOption
from someipy._internal._sd.options.multicast import (
    IpV4MulticastOption,
    IpV6MulticastOption,
)
from someipy._internal._sd.options.sd_endpoint import (
    IpV4SdEndpointOption,
    IpV6SdEndpointOption,
)
from someipy._internal.transport_layer_protocol import TransportLayerProtocol
from someipy._internal.utils import is_bit_set
from someipy._internal._sd.sd_message import SdMessage

SERVICE_ID_SD = 0xFFFF
METHOD_ID_SD = 0x8100
CLIENT_ID_SD = 0x0000
PROTOCOL_VERSION_SD = 0x01
INTERFACE_VERSION_SD = 0x01
MESSAGE_TYPE_SD = 0x02
RETURN_CODE_SD = 0x00

MINIMAL_HEADER_SIZE = 16


class SdOptionOnWireType(Enum):
    CONFIGURATION = 0x01
    LOAD_BALANCING = 0x02
    IPV4_ENDPOINT = 0x04
    IPV6_ENDPOINT = 0x06
    IPV4_MULTICAST = 0x14
    IPV6_MULTICAST = 0x16
    IPV4_SD_ENDPOINT = 0x24
    IPV6_SD_ENDPOINT = 0x26


class SdEntryOnWireType(Enum):
    FIND_SERVICE = 0x00
    OFFER_SERVICE = 0x01
    STOP_OFFER_SERVICE = 0x01  # with TTL to 0x000000
    SUBSCRIBE_EVENT_GROUP = 0x06
    STOP_SUBSCRIBE_EVENT_GROUP = 0x06  # with TTL to 0x000000
    SUBSCRIBE_EVENT_GROUP_ACK = 0x07
    SUBSCRIBE_EVENT_GROUP_NACK = 0x07  # with TTL to 0x000000


def is_sd_message(data: bytes) -> bool:
    if len(data) < MINIMAL_HEADER_SIZE:
        return False

    service_id, method_id, length = struct.unpack(">HHI", data[0:8])

    if length < 12:
        return False

    if length > len(data) - 8:
        return False

    (
        client_id,
        session_id,
        protocol_version,
        interface_version,
        message_type,
        return_code,
    ) = struct.unpack(">HHBBBB", data[8:16])

    return (
        service_id == SERVICE_ID_SD
        and method_id == METHOD_ID_SD
        and client_id == CLIENT_ID_SD
        and protocol_version == PROTOCOL_VERSION_SD
        and interface_version == INTERFACE_VERSION_SD
        and message_type == MESSAGE_TYPE_SD
        and return_code == RETURN_CODE_SD
        and session_id != 0
    )


@dataclass
class CommonEntryData:
    type_field_value: int
    index_first_option: int
    index_second_option: int
    num_options_1: int
    num_options_2: int
    service_id: int
    instance_id: int
    major_version: int
    ttl: int


@dataclass
class CommonOptionData:
    option_length: int
    option_type: SdOptionOnWireType
    discardable_flag: bool


def deserialize_common_entry_data(data: bytes) -> CommonEntryData:
    type_field_value, index_first_option, index_second_option = struct.unpack(
        ">BBB", data[0:3]
    )

    num_options_1 = struct.unpack(">B", data[3:4])[0]  # higher 4 bits
    num_options_1 = (num_options_1 >> 4) & 0x0F

    num_options_2 = struct.unpack(">B", data[3:4])[0]  # lower 4 bits
    num_options_2 = num_options_2 & 0x0F

    service_id, instance_id, major_version = struct.unpack(">HHB", data[4:9])
    (ttl,) = struct.unpack(">I", data[8:12])
    ttl = ttl & 0xFFFFFF

    return CommonEntryData(
        type_field_value,
        index_first_option,
        index_second_option,
        num_options_1,
        num_options_2,
        service_id,
        instance_id,
        major_version,
        ttl,
    )


def deserialize_common_option_data(data: bytes) -> CommonOptionData:
    option_length, option_type, discardable_flag_value = struct.unpack(
        ">HBB", data[0:4]
    )
    option_type = SdOptionOnWireType(option_type)
    discardable_flag = is_bit_set(discardable_flag_value, 7)
    return CommonOptionData(option_length, option_type, discardable_flag)


def deserialize_ipv4_endpoint_option(data: bytes) -> IpV4EndpointOption:
    ip1, ip2, ip3, ip4, _, protocol_value, port = struct.unpack(">BBBBBBH", data[0:8])
    address = ipaddress.IPv4Address(f"{ip1}.{ip2}.{ip3}.{ip4}")
    protocol = TransportLayerProtocol(protocol_value)
    return IpV4EndpointOption(address=address, protocol=protocol, port=port)


def deserialize_ipv6_endpoint_option(data: bytes) -> IpV6EndpointOption:
    packed = data[0:16]
    ipv6_str = socket.inet_ntop(socket.AF_INET6, packed)
    address = ipaddress.IPv6Address(ipv6_str)
    _, protocol_value, port = struct.unpack(">BBH", data[16:20])
    protocol = TransportLayerProtocol(protocol_value)
    return IpV6EndpointOption(address=address, protocol=protocol, port=port)


def deserialize_ipv4_multicast_option(data: bytes) -> IpV4MulticastOption:
    ip1, ip2, ip3, ip4, _, protocol_value, port = struct.unpack(">BBBBBBH", data[0:8])
    address = ipaddress.IPv4Address(f"{ip1}.{ip2}.{ip3}.{ip4}")
    protocol = TransportLayerProtocol(protocol_value)
    return IpV4MulticastOption(address=address, protocol=protocol, port=port)


def deserialize_ipv6_multicast_option(data: bytes) -> IpV6MulticastOption:
    packed = data[0:16]
    ipv6_str = socket.inet_ntop(socket.AF_INET6, packed)
    address = ipaddress.IPv6Address(ipv6_str)
    _, protocol_value, port = struct.unpack(">BBH", data[16:20])
    protocol = TransportLayerProtocol(protocol_value)
    return IpV6MulticastOption(address=address, protocol=protocol, port=port)


def deserialize_ipv4_sd_endpoint_option(data: bytes) -> IpV4SdEndpointOption:
    ip1, ip2, ip3, ip4, _, _, port = struct.unpack(">BBBBBBH", data[0:8])
    address = ipaddress.IPv4Address(f"{ip1}.{ip2}.{ip3}.{ip4}")
    return IpV4SdEndpointOption(address=address, port=port)


def deserialize_ipv6_sd_endpoint_option(data: bytes) -> IpV6SdEndpointOption:
    packed = data[0:16]
    ipv6_str = socket.inet_ntop(socket.AF_INET6, packed)
    address = ipaddress.IPv6Address(ipv6_str)
    _, _, port = struct.unpack(">BBH", data[16:20])
    return IpV6SdEndpointOption(address=address, port=port)


def deserialize_load_balancing_option(data: bytes) -> LoadBalancingOption:
    priority, weight = struct.unpack(">HH", data[0:4])
    return LoadBalancingOption(priority=priority, weight=weight)


def deserialize_sd_message(
    data: bytes, source: str, port: int, multicast: bool
) -> SdMessage:

    if not is_sd_message(data):
        raise ValueError("The provided data is not a valid SOME/IP-SD message.")

    service_id, method_id, length = struct.unpack(">HHI", data[0:8])
    if length <= 0:
        raise ValueError(f"Length in SOME/IP header is <=0 ({length})")

    if length < 8:
        raise ValueError(f"Length in SOME/IP header is <8 ({length})")

    (
        client_id,
        session_id,
        protocol_version,
        interface_version,
        message_type,
        return_code,
    ) = struct.unpack(">HHBBBB", data[8:16])

    (flags,) = struct.unpack(">B", data[16:17])
    reboot_flag = is_bit_set(flags, 7)
    unicast_flag = is_bit_set(flags, 6)

    # Constants for byte positions inside the SD header
    SD_POSITION_ENTRY_LENGTH = 20
    SD_START_POSITION_ENTRIES = 24

    # Constants for length of sections in the SD header
    SD_SINGLE_ENTRY_LENGTH_BYTES = 16

    (length_entries,) = struct.unpack(
        ">I", data[SD_POSITION_ENTRY_LENGTH : SD_POSITION_ENTRY_LENGTH + 4]
    )

    number_of_entries = int(length_entries / SD_SINGLE_ENTRY_LENGTH_BYTES)

    pos_length_options = SD_POSITION_ENTRY_LENGTH + 4 + length_entries
    (length_options,) = struct.unpack(
        ">I", data[pos_length_options : pos_length_options + 4]
    )
    pos_start_options = pos_length_options + 4

    current_pos_option = pos_start_options
    bytes_options_left = length_options

    options = []
    while bytes_options_left > 0:

        common_option_data = deserialize_common_option_data(
            data[current_pos_option : current_pos_option + 4]
        )

        option_type = common_option_data.option_type

        if option_type == SdOptionOnWireType.IPV4_ENDPOINT:
            sd_option = deserialize_ipv4_endpoint_option(
                data[current_pos_option + 4 : current_pos_option + 12]
            )
            options.append(sd_option)

        elif option_type == SdOptionOnWireType.IPV6_ENDPOINT:
            sd_option = deserialize_ipv6_endpoint_option(
                data[current_pos_option + 4 : current_pos_option + 24]
            )
            options.append(sd_option)
        elif option_type == SdOptionOnWireType.IPV4_MULTICAST:
            sd_option = deserialize_ipv4_multicast_option(
                data[current_pos_option + 4 : current_pos_option + 12]
            )
            options.append(sd_option)
        elif option_type == SdOptionOnWireType.IPV6_MULTICAST:
            sd_option = deserialize_ipv6_multicast_option(
                data[current_pos_option + 4 : current_pos_option + 24]
            )
            options.append(sd_option)
        elif option_type == SdOptionOnWireType.IPV4_SD_ENDPOINT:
            sd_option = deserialize_ipv4_sd_endpoint_option(
                data[current_pos_option + 4 : current_pos_option + 12]
            )
            options.append(sd_option)
        elif option_type == SdOptionOnWireType.IPV6_SD_ENDPOINT:
            sd_option = deserialize_ipv6_sd_endpoint_option(
                data[current_pos_option + 4 : current_pos_option + 24]
            )
            options.append(sd_option)
        elif option_type == SdOptionOnWireType.CONFIGURATION:
            dummy_config_option = ConfigurationOption()
            options.append(dummy_config_option)
        elif option_type == SdOptionOnWireType.LOAD_BALANCING:
            sd_option = deserialize_load_balancing_option(
                data[current_pos_option + 4 : current_pos_option + 8]
            )
            options.append(sd_option)

        # Subtract 3 bytes first for length and type
        bytes_options_left -= common_option_data.option_length + 3
        current_pos_option += common_option_data.option_length + 3

    # Read in all Service and Event Group entries
    entries = []
    for i in range(number_of_entries):
        start_entry = SD_START_POSITION_ENTRIES + (i * SD_SINGLE_ENTRY_LENGTH_BYTES)
        end_entry = start_entry + SD_SINGLE_ENTRY_LENGTH_BYTES

        common_entry_data = deserialize_common_entry_data(
            data[start_entry : start_entry + 12]
        )

        if (
            common_entry_data.type_field_value == SdEntryOnWireType.OFFER_SERVICE.value
            and common_entry_data.ttl != 0
        ):
            (minor_version,) = struct.unpack(
                ">I", data[start_entry + 12 : start_entry + 16]
            )

            applicable_options = []
            for j in range(common_entry_data.num_options_1):
                applicable_options.append(
                    options[common_entry_data.index_first_option + j]
                )
            for j in range(common_entry_data.num_options_2):
                applicable_options.append(
                    options[common_entry_data.index_second_option + j]
                )

            offer_service_entry = OfferServiceEntry(
                service_id=common_entry_data.service_id,
                instance_id=common_entry_data.instance_id,
                major_version=common_entry_data.major_version,
                minor_version=minor_version,
                ttl=common_entry_data.ttl,
                ip_v4_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV4EndpointOption)
                ],
                ip_v6_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV6EndpointOption)
                ],
            )
            entries.append(offer_service_entry)

        elif (
            common_entry_data.type_field_value
            == SdEntryOnWireType.STOP_OFFER_SERVICE.value
            and common_entry_data.ttl == 0
        ):
            (minor_version,) = struct.unpack(
                ">I", data[start_entry + 12 : start_entry + 16]
            )

            applicable_options = []
            for j in range(common_entry_data.num_options_1):
                applicable_options.append(
                    options[common_entry_data.index_first_option + j]
                )
            for j in range(common_entry_data.num_options_2):
                applicable_options.append(
                    options[common_entry_data.index_second_option + j]
                )

            stop_offer_service_entry = StopOfferServiceEntry(
                service_id=common_entry_data.service_id,
                instance_id=common_entry_data.instance_id,
                major_version=common_entry_data.major_version,
                minor_version=minor_version,
                ip_v4_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV4EndpointOption)
                ],
                ip_v6_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV6EndpointOption)
                ],
            )
            entries.append(stop_offer_service_entry)

        elif common_entry_data.type_field_value == SdEntryOnWireType.FIND_SERVICE.value:
            (minor_version,) = struct.unpack(
                ">I", data[start_entry + 12 : start_entry + 16]
            )

            find_service_entry = FindServiceEntry(
                service_id=common_entry_data.service_id,
                instance_id=common_entry_data.instance_id,
                major_version=common_entry_data.major_version,
                minor_version=minor_version,
            )
            entries.append(find_service_entry)

        elif (
            common_entry_data.type_field_value
            == SdEntryOnWireType.SUBSCRIBE_EVENT_GROUP.value
            and common_entry_data.ttl != 0
        ):
            initial_data_requested_flag_counter_value, eventgroup_id = struct.unpack(
                ">BH", data[start_entry + 13 : start_entry + 16]
            )
            initial_data_requested_flag = is_bit_set(
                initial_data_requested_flag_counter_value, 7
            )
            counter = initial_data_requested_flag_counter_value & 0xF

            applicable_options = []
            for j in range(common_entry_data.num_options_1):
                applicable_options.append(
                    options[common_entry_data.index_first_option + j]
                )
            for j in range(common_entry_data.num_options_2):
                applicable_options.append(
                    options[common_entry_data.index_second_option + j]
                )

            subscribe_eventgroup_entry = SubscribeEventGroupEntry(
                service_id=common_entry_data.service_id,
                instance_id=common_entry_data.instance_id,
                major_version=common_entry_data.major_version,
                minor_version=minor_version,
                ttl=common_entry_data.ttl,
                eventgroup_id=eventgroup_id,
                counter=counter,
                ip_v4_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV4EndpointOption)
                ],
                ip_v6_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV6EndpointOption)
                ],
            )
            entries.append(subscribe_eventgroup_entry)

        elif (
            common_entry_data.type_field_value
            == SdEntryOnWireType.STOP_SUBSCRIBE_EVENT_GROUP.value
            and common_entry_data.ttl == 0
        ):
            initial_data_requested_flag_counter_value, eventgroup_id = struct.unpack(
                ">BH", data[start_entry + 13 : start_entry + 16]
            )
            initial_data_requested_flag = is_bit_set(
                initial_data_requested_flag_counter_value, 7
            )
            counter = initial_data_requested_flag_counter_value & 0xF

            applicable_options = []
            for j in range(common_entry_data.num_options_1):
                applicable_options.append(
                    options[common_entry_data.index_first_option + j]
                )
            for j in range(common_entry_data.num_options_2):
                applicable_options.append(
                    options[common_entry_data.index_second_option + j]
                )

            stop_subscribe_eventgroup_entry = StopSubscribeEventGroupEntry(
                service_id=common_entry_data.service_id,
                instance_id=common_entry_data.instance_id,
                major_version=common_entry_data.major_version,
                minor_version=minor_version,
                ttl=common_entry_data.ttl,
                ip_v4_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV4EndpointOption)
                ],
                ip_v6_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV6EndpointOption)
                ],
            )
            entries.append(stop_subscribe_eventgroup_entry)

        elif (
            common_entry_data.type_field_value
            == SdEntryOnWireType.SUBSCRIBE_EVENT_GROUP_ACK.value
            and common_entry_data.ttl != 0
        ):
            initial_data_requested_flag_counter_value, eventgroup_id = struct.unpack(
                ">BH", data[start_entry + 13 : start_entry + 16]
            )
            initial_data_requested_flag = is_bit_set(
                initial_data_requested_flag_counter_value, 7
            )
            counter = initial_data_requested_flag_counter_value & 0xF

            applicable_options = []
            for j in range(common_entry_data.num_options_1):
                applicable_options.append(
                    options[common_entry_data.index_first_option + j]
                )
            for j in range(common_entry_data.num_options_2):
                applicable_options.append(
                    options[common_entry_data.index_second_option + j]
                )

            subscribe_ack_eventgroup_entry = SubscribeAckEventGroupEntry(
                service_id=common_entry_data.service_id,
                instance_id=common_entry_data.instance_id,
                major_version=common_entry_data.major_version,
                minor_version=minor_version,
                ttl=common_entry_data.ttl,
                eventgroup_id=eventgroup_id,
                counter=counter,
                ip_v4_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV4EndpointOption)
                ],
                ip_v6_endpoints=[
                    o for o in applicable_options if isinstance(o, IpV6EndpointOption)
                ],
            )
            entries.append(subscribe_ack_eventgroup_entry)

        elif (
            common_entry_data.type_field_value
            == SdEntryOnWireType.SUBSCRIBE_EVENT_GROUP_NACK.value
            and common_entry_data.ttl == 0
        ):
            initial_data_requested_flag_counter_value, eventgroup_id = struct.unpack(
                ">BH", data[start_entry + 13 : start_entry + 16]
            )
            initial_data_requested_flag = is_bit_set(
                initial_data_requested_flag_counter_value, 7
            )
            counter = initial_data_requested_flag_counter_value & 0xF

            applicable_options = []
            for j in range(common_entry_data.num_options_1):
                applicable_options.append(
                    options[common_entry_data.index_first_option + j]
                )
            for j in range(common_entry_data.num_options_2):
                applicable_options.append(
                    options[common_entry_data.index_second_option + j]
                )

            subscribe_ack_eventgroup_entry = SubscribeEventGroupNackEntry(
                service_id=common_entry_data.service_id,
                instance_id=common_entry_data.instance_id,
                major_version=common_entry_data.major_version,
                minor_version=minor_version,
                eventgroup_id=eventgroup_id,
                counter=counter,
            )
            entries.append(subscribe_ack_eventgroup_entry)

    sd_message = SdMessage()
    sd_message.source = source
    sd_message.source_port = port
    sd_message.multicast = multicast
    sd_message.session_id = session_id
    sd_message.reboot_flag = reboot_flag
    sd_message.entries = entries
    return sd_message



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, List, TypeVar, TypedDict, get_type_hints


class BaseMessage(TypedDict):
    type: str


class OfferServiceRequest(BaseMessage):
    service_id: int
    instance_id: int
    major_version: int
    minor_version: int
    endpoint_ip: str
    endpoint_port: int
    ttl: int
    eventgroup_list: List[str]
    method_list: List[str]
    cyclic_offer_delay_ms: int


class StopOfferServiceRequest(BaseMessage):
    service_id: int
    instance_id: int
    major_version: int
    minor_version: int
    endpoint_ip: str
    endpoint_port: int
    ttl: int
    eventgroup_list: List[str]
    method_list: List[str]
    cyclic_offer_delay_ms: int


class FindServiceRequest(BaseMessage):
    service_id: int
    instance_id: int
    major_version: int
    minor_version: int


class FindServiceResponse(BaseMessage):
    success: bool
    service_id: int
    instance_id: int
    major_version: int
    minor_version: int
    endpoint_ip: str
    endpoint_port: int


class OutboundCallMethodRequest(BaseMessage):
    service_id: int
    instance_id: int
    method_id: int
    client_id: int
    session_id: int
    protocol_version: int
    major_version: int
    minor_version: int
    dst_endpoint_ip: str
    dst_endpoint_port: int
    src_endpoint_ip: str
    src_endpoint_port: int
    protocol: int
    payload: str


class OutboundCallMethodResponse(BaseMessage):
    service_id: int
    method_id: int
    client_id: int
    session_id: int
    return_code: int
    dst_endpoint_ip: str
    dst_endpoint_port: int
    payload: str


class InboundCallMethodRequest(BaseMessage):
    service_id: int
    instance_id: int
    method_id: int
    client_id: int
    session_id: int
    protocol_version: int
    interface_version: int
    major_version: int
    minor_version: int
    message_type: int
    src_endpoint_ip: str
    src_endpoint_port: int
    protocol: int
    payload: str


class InboundCallMethodResponse(BaseMessage):
    service_id: int
    instance_id: int
    method_id: int
    client_id: int
    session_id: int
    protocol_version: int
    interface_version: int
    major_version: int
    minor_version: int
    message_type: int
    src_endpoint_ip: str
    src_endpoint_port: int
    protocol: int
    payload: str
    return_code: int


class SendEventRequest(BaseMessage):
    service_id: int
    instance_id: int
    major_version: int
    client_id: int
    session_id: int
    eventgroup_id: int
    event: str  # Serialized event object
    src_endpoint_ip: str
    src_endpoint_port: int
    payload: str


class SubscribeEventGroupRequest(BaseMessage):
    service_id: int
    instance_id: int
    major_version: int
    ttl_subscription: int
    eventgroup: str
    client_endpoint_ip: str
    client_endpoint_port: int
    udp: bool
    tcp: bool


class StopSubscribeEventGroupRequest(BaseMessage):
    service_id: int
    instance_id: int
    major_version: int
    eventgroup: str
    client_endpoint_ip: str
    client_endpoint_port: int
    udp: bool
    tcp: bool


class ReceivedEvent(BaseMessage):
    service_id: int
    event_id: int
    src_endpoint_ip: str
    src_endpoint_port: str
    payload: str


T = TypeVar("T", bound=BaseMessage)


def create_uds_message(message_python_type: type[T], **kwargs: Any) -> T:
    type_value = message_python_type.__name__
    message = {"type": type_value, **kwargs}

    # For development: Test if all keys are given using type inspection
    if set(message.keys()) != set(get_type_hints(message_python_type).keys()):
        raise RuntimeError(
            f"Invalid data passed to create_uds_message for type {type_value}"
        )

    return message



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import time
from typing import List
from someipy._internal.utils import EndpointType


class EventGroupSubscriber:
    eventgroup_id: int
    endpoint: EndpointType
    ttl: int
    last_ts_ms: int

    def __init__(self, eventgroup_id: int, endpoint: EndpointType, ttl: int):
        self.eventgroup_id = eventgroup_id
        self.endpoint = endpoint
        self.ttl = ttl
        self.last_ts_ms = int(time.time() * 1000)

    def __eq__(self, __value: object) -> bool:
        return (
            self.eventgroup_id == __value.eventgroup_id
            and self.endpoint[0] == __value.endpoint[0]
            and self.endpoint[1] == __value.endpoint[1]
        )


class Subscribers:
    def __init__(self):
        self._subscribers: List[EventGroupSubscriber] = []

    def update(self):
        # From SOME/IP specification:
        # TTL shall be set to the lifetime of the subscription.
        # If set to 0xFFFFFF, the Subscribe Eventgroup entry shall be considered valid
        # until the next reboot.
        time_now_ms = int(time.time() * 1000)

        self._subscribers = [
            s
            for s in self._subscribers
            if (s.ttl == 0xFFFFFF) or (time_now_ms < (s.last_ts_ms + (s.ttl * 1000.0)))
        ]

    def add_subscriber(self, new_subscriber: EventGroupSubscriber) -> None:
        time_now_ms = int(time.time() * 1000.0)
        for s in self._subscribers:
            if new_subscriber == s:
                s.last_ts_ms = time_now_ms
                return
        new_subscriber.last_ts_ms = time_now_ms
        self._subscribers.append(new_subscriber)

    def remove_subscriber(self, subscriber: EventGroupSubscriber) -> None:
        if subscriber in self._subscribers:
            self._subscribers.remove(subscriber)

    def clear(self) -> None:
        self._subscribers = []

    @property
    def subscribers(self) -> List[EventGroupSubscriber]:
        return self._subscribers

    @property
    def has_subscribers(self) -> bool:
        return len(self._subscribers) > 0



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import sys
from someipy.someipy_logging import get_someipy_log_level


def setup_console_handler(
    formatter: logging.Formatter, level: int
) -> logging.StreamHandler:
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(level)
    console_handler.setFormatter(formatter)
    return console_handler


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(f"someipy.{name}")
    log_level = get_someipy_log_level()

    formatter = logging.Formatter(
        "%(asctime)s.%(msecs)03d %(name)s [%(levelname)s]: %(message)s",
        datefmt="%Y-%m-%d,%H:%M:%S",
    )

    # Check if the logger already has a console handler
    if not any(
        isinstance(handler, logging.StreamHandler) for handler in logger.handlers
    ):
        console_handler = setup_console_handler(formatter, log_level)
        logger.addHandler(console_handler)

    logger.setLevel(log_level)

    return logger



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from someipy._internal._common.endpoint import Endpoint
from someipy._internal.transport_layer_protocol import TransportLayerProtocol
from someipy.service import EventGroup


class Subscription:
    def __init__(
        self,
        service_id: int,
        instance_id: int,
        major_version: int,
        eventgroup: EventGroup,
        ttl_seconds: int,
        client_endpoint: Endpoint,
        server_endpoint: Endpoint,
        protocols: frozenset[TransportLayerProtocol],
        timestamp_last_update: float = 0.0,
    ):
        self.service_id = service_id
        self.instance_id = instance_id
        self.major_version = major_version
        self.eventgroup = eventgroup
        self.ttl_seconds = ttl_seconds

        self.client_endpoint = client_endpoint
        self.server_endpoint = server_endpoint
        self.protocols = protocols
        self.timestamp_last_update = timestamp_last_update

    def __eq__(self, value: "Subscription") -> bool:
        return (
            self.service_id == value.service_id
            and self.instance_id == value.instance_id
            and self.major_version == value.major_version
            and self.eventgroup == value.eventgroup
            and self.client_endpoint == value.client_endpoint
            and self.server_endpoint == value.server_endpoint
            and self.protocols == value.protocols
        )

    def __hash__(self) -> int:
        # Do not include the timestamp and ttl in the hash calculation
        return hash(
            (
                self.service_id,
                self.instance_id,
                self.major_version,
                self.eventgroup,
                self.client_endpoint,
                self.server_endpoint,
                self.protocols,
            )
        )



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Tuple

from someipy._internal._daemon.uds_messages import (
    InboundCallMethodResponse,
    ReceivedEvent,
)
from someipy.service import Service


class ClientInstanceInterface(ABC):

    @abstractmethod
    def _method_call_data_received(self, message: InboundCallMethodResponse) -> None:
        pass

    @abstractmethod
    def _event_data_received(self, message: ReceivedEvent) -> None:
        pass

    @property
    @abstractmethod
    def service(self) -> Service:
        pass

    @property
    @abstractmethod
    def instance_id(self) -> int:
        pass

    @property
    @abstractmethod
    def endpoint(self) -> Tuple[str, int]:
        pass


class ServerInstanceInterface(ABC):

    @property
    @abstractmethod
    def service(self) -> Service:
        pass

    @property
    @abstractmethod
    def instance_id(self) -> int:
        pass

    @property
    @abstractmethod
    def endpoint_ip(self) -> str:
        pass

    @property
    @abstractmethod
    def endpoint_port(self) -> int:
        pass



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from abc import ABC, abstractmethod, abstractproperty
from logging import Logger
from typing import Callable, Tuple, Any, Union
from someipy._internal.logging import get_logger
from someipy._internal.someip_message import SomeIpMessage
from someipy._internal.tcp_client_manager import (
    TcpClientManagerInterface,
    TcpClientProtocolInterface,
)
from someipy._internal.tcp_connection import TcpConnection
from someipy._internal.transport_layer_protocol import TransportLayerProtocol
from someipy._internal.utils import EndpointType
from someipy._internal.someip_data_processor import SomeipDataProcessor


class SomeipEndpoint(ABC):
    """
    A class representing an endpoint (UDP or TCP) which fires a callback when a SomeIpMessage is received.
    It can also send messages to a specific endpoint or broadcast to all connected endpoints (for UDP multicast or to all connected
    TCP clients).
    """

    @abstractmethod
    def set_someip_callback(
        self,
        callback_func: Callable[
            [SomeIpMessage, Tuple[str, int], Tuple[str, int], TransportLayerProtocol],
            None,
        ],
    ) -> None:
        pass

    @abstractmethod
    def sendto(self, data: bytes, addr: EndpointType) -> None:
        pass

    @abstractmethod
    def sendtoall(self, data: bytes) -> None:
        pass

    @abstractmethod
    def shutdown(self) -> None:
        pass

    @abstractmethod
    def protocol(self) -> TransportLayerProtocol:
        pass

    @abstractmethod
    def src_ip(self) -> str:
        pass

    @abstractmethod
    def src_port(self) -> int:
        pass

    @abstractmethod
    def dst_ip(self) -> str:
        pass

    @abstractmethod
    def dst_port(self) -> int:
        pass


class UDPSomeipEndpoint(SomeipEndpoint, asyncio.DatagramProtocol):
    def __init__(self, ip: str, port: int):
        self._callback: Callable[
            [SomeIpMessage, Tuple[str, int], Tuple[str, int], TransportLayerProtocol],
            None,
        ] = None
        self._transport = None
        self._processor = SomeipDataProcessor()
        self._protocol = TransportLayerProtocol.UDP

        self._ip = ip
        self._port = port

    def set_someip_callback(
        self,
        callback_func: Callable[
            [SomeIpMessage, Tuple[str, int], Tuple[str, int], TransportLayerProtocol],
            None,
        ],
    ) -> None:
        self._callback = callback_func

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
        self._transport = transport

    def connection_lost(self, exc: Exception) -> None:
        pass

    def datagram_received(self, data: bytes, addr: Tuple[Union[str, Any, int]]) -> None:
        result = self._processor.process_data(data)
        if result and self._callback is not None:
            self._callback(
                self._processor.someip_message,
                addr,
                (self._ip, self._port),
                self._protocol,
            )

    def sendto(self, data: bytes, addr: EndpointType) -> None:
        if self._transport is not None:
            self._transport.sendto(data, (str(addr[0]), addr[1]))

    def sendtoall(self, data: bytes) -> None:
        # TODO: Implement for multicast support
        raise NotImplementedError("No implementation for UDP yet")

    def shutdown(self) -> None:
        if self._transport is not None:
            self._transport.close()

    def protocol(self) -> TransportLayerProtocol:
        return self._protocol

    def src_ip(self) -> str:
        return self._ip

    def src_port(self) -> int:
        return self._port

    def dst_ip(self) -> str:
        return 0

    def dst_port(self) -> int:
        return 0


class TCPSomeipEndpoint(SomeipEndpoint):
    def __init__(self, server, manager, ip: str, port: int):
        self._server: asyncio.Server = server
        self._manager: TcpClientManagerInterface = manager
        self._protocol: TransportLayerProtocol = TransportLayerProtocol.TCP
        self._ip = ip
        self._port = port

    def set_someip_callback(
        self,
        callback_func: Callable[
            [SomeIpMessage, Tuple[str, int], Tuple[str, int], TransportLayerProtocol],
            None,
        ],
    ) -> None:
        self._manager.register_callback(callback_func)

    def sendto(self, data: bytes, addr: EndpointType) -> None:
        if self._manager is None:
            return
        client: TcpClientProtocolInterface = self._manager.get_client(addr[0], addr[1])
        if client is not None:
            client.write(data)

    def sendtoall(self, data: bytes) -> None:
        if self._manager is None:
            return
        for c in self._manager.get_all_clients():
            c.write(data)

    def shutdown(self) -> None:
        if self._manager is None:
            return
        self._server.close()

    def protocol(self) -> TransportLayerProtocol:
        return self._protocol

    def src_ip(self) -> str:
        return self._ip

    def src_port(self) -> int:
        return self._port

    def dst_ip(self) -> str:
        return 0

    def dst_port(self) -> int:
        return 0


class TCPClientSomeipEndpoint(SomeipEndpoint):
    def __init__(
        self,
        dst_ip: str,
        dst_port: str,
        src_ip: str,
        src_port: str,
        logger: Logger = None,
    ):
        self._tcp_connection: TcpConnection = None
        self._protocol: TransportLayerProtocol = TransportLayerProtocol.TCP
        self._dst_ip = dst_ip
        self._dst_port = dst_port
        self._src_ip = src_ip
        self._src_port = src_port
        self._logger = logger

        self._callback: Callable[
            [SomeIpMessage, Tuple[str, int], Tuple[str, int], TransportLayerProtocol],
            None,
        ] = None

        self._connect_and_receive_task: asyncio.Task = asyncio.create_task(
            self._connect_and_receive()
        )

    def _log(self, message: str) -> None:
        if self._logger:
            self._logger.debug(message)

    def is_connected(self) -> bool:
        return self._tcp_connection is not None and self._tcp_connection.is_open()

    async def _connect_and_receive(self) -> None:
        try:
            while True:
                self._log(f"Trying to connect to ({self._dst_ip}, {self._dst_port})")

                self._tcp_connection = TcpConnection(self._dst_ip, self._dst_port)

                try:
                    await self._tcp_connection.connect(self._src_ip, self._src_port)
                except OSError:
                    self._log(
                        f"Connection refused while connecting to ({self._dst_ip}, {self._dst_port}). Try to reconnect in 1 second"
                    )
                    # Wait a second before trying to connect again
                    await asyncio.sleep(1.0)
                    continue

                self._log(f"Start reading on port {self._src_port}")

                someip_processor = SomeipDataProcessor()

                while self._tcp_connection.is_open():
                    try:
                        new_data = await asyncio.wait_for(
                            self._tcp_connection.reader.read(
                                someip_processor.expected_bytes
                            ),
                            3.0,
                        )

                        if someip_processor.process_data(new_data):
                            self._callback(
                                someip_processor.someip_message,
                                (self._dst_ip, self._dst_port),
                                (self._src_ip, self._src_port),
                                self._protocol,
                            )

                    except asyncio.TimeoutError:
                        self._log(
                            f"Timeout reading from TCP connection ({self._src_ip}, {self._src_port})"
                        )
                    except Exception as e:
                        self._log(
                            f"Error while reading from TCP connection ({self._src_ip}, {self._src_port}): {e}"
                        )

                await self._tcp_connection.close()

        except asyncio.CancelledError:
            if self._tcp_connection.is_open():
                await self._tcp_connection.close()
            self._log("TCP task is cancelled. Raise again.")
            raise

    def set_someip_callback(
        self,
        callback_func: Callable[
            [SomeIpMessage, Tuple[str, int], Tuple[str, int], TransportLayerProtocol],
            None,
        ],
    ) -> None:
        self._callback = callback_func

    def sendto(self, data: bytes, addr: EndpointType) -> None:
        if self._tcp_connection is not None and self._tcp_connection.is_open():
            self._tcp_connection.writer.write(data)

    def sendtoall(self, data: bytes) -> None:
        raise NotImplementedError("No implementation for TCP client endpoint.")

    def shutdown(self) -> None:
        try:
            if self._connect_and_receive_task:
                self._connect_and_receive_task.cancel()
        except asyncio.CancelledError:
            self._log("Shutdown task was cancelled.")

    def protocol(self) -> TransportLayerProtocol:
        return self._protocol

    def src_ip(self) -> str:
        return self._src_ip

    def src_port(self) -> int:
        return self._src_port

    def dst_ip(self) -> str:
        return self._dst_ip

    def dst_port(self) -> int:
        return self._dst_port



from dataclasses import dataclass

import pytest
from someipy.serialization import (
    Uint8,
    Uint16,
    Uint32,
    Uint64,
    Sint8,
    Sint16,
    Sint32,
    Sint64,
    Bool,
    Float32,
    Float64,
    SomeIpPayload,
    SomeIpFixedSizeArray,
    SomeIpDynamicSizeArray,
    SomeIpFixedSizeString,
    SomeIpDynamicSizeString,
)


def test_base_types_len():
    assert 1 == len(Uint8(1))
    assert 2 == len(Uint16(1))
    assert 4 == len(Uint32(1))
    assert 8 == len(Uint64(1))

    assert 1 == len(Sint8(-1))
    assert 2 == len(Sint16(-1))
    assert 4 == len(Sint32(-1))
    assert 8 == len(Sint64(-1))

    assert 1 == len(Bool(True))
    assert 4 == len(Float32(1.0))
    assert 8 == len(Float64(1.0))


@dataclass
class MsgBaseTypesOnly(SomeIpPayload):
    """
    The dataclass decorator will generate an __eq__ method which
    can be used for comparing the content of two messages
    """

    # Expected 13 bytes in total
    x: Uint8  # 1 byte
    y: Uint32  # 4 bytes
    z: Float64  # 8 bytes

    def __init__(self):
        self.x = Uint8(0)
        self.y = Uint32(0)
        self.z = Float64(0.0)


@dataclass
class MsgWithOneStruct(SomeIpPayload):
    # Expected 1 + 13 + 4 bytes in total
    a: Uint8
    b: MsgBaseTypesOnly
    c: Sint32

    def __init__(self):
        self.a = Uint8(0)
        self.b = MsgBaseTypesOnly()
        self.c = Sint32(0)


def test_someip_payload_len():
    assert 13 == len(MsgBaseTypesOnly())
    assert (13 + 1 + 4) == len(MsgWithOneStruct())


def test_struct_equals_operator():
    m_1 = MsgBaseTypesOnly()
    m_2 = MsgBaseTypesOnly()
    assert m_1 == m_2

    m_1.x = Uint8(1)
    assert m_1 != m_2

    m_2.x = Uint8(1)
    assert m_1 == m_2


def test_struct_serialization_and_deserialization():
    m = MsgBaseTypesOnly()
    m.y = Uint32(4)
    # Expected: 0x 00 00000004 0000000000000000
    assert bytes.fromhex("00000000040000000000000000") == m.serialize()

    m.x = Uint8(255)
    # Expected: 0x FF 00000004 0000000000000000
    assert bytes.fromhex("ff000000040000000000000000") == m.serialize()

    m_again = MsgBaseTypesOnly().deserialize(
        bytes.fromhex("ff000000040000000000000000")
    )
    assert m_again == m

    # The first byte is changed now to decimal 254
    m_different = MsgBaseTypesOnly().deserialize(
        bytes.fromhex("fe000000040000000000000000")
    )
    assert m_different != m_again
    assert m_different != m


def test_fixed_size_array_length():
    a = SomeIpFixedSizeArray(Uint16, 10)
    expected_bytes = len(Uint16()) * 10
    assert len(a) == expected_bytes


def test_fixed_size_array_equals_operator():
    a = SomeIpFixedSizeArray(Uint16, 10)
    b = SomeIpFixedSizeArray(Uint16, 10)
    assert a == b

    c = SomeIpFixedSizeArray(Uint8, 10)
    assert a != c

    a.data[0] = Uint16(5)
    assert a != b

    b.data[0] = Uint16(5)
    assert a == b


def test_fixed_size_array_serialization_and_deserialization():
    a = SomeIpFixedSizeArray(Uint8, 4)
    # Expected hex: 0x 00 00 00 00
    assert bytes.fromhex("00000000") == a.serialize()

    a.data[0] = Uint8(1)
    a.data[1] = Uint8(2)
    a.data[2] = Uint8(3)
    a.data[3] = Uint8(4)
    # Expected hex: 0x 01 02 03 04
    assert bytes.fromhex("01020304") == a.serialize()

    a_again = SomeIpFixedSizeArray(Uint8, 4).deserialize(bytes.fromhex("01020304"))
    assert a_again == a


def test_dynamic_size_array_length():
    a = SomeIpDynamicSizeArray(Uint16)

    assert len(a) == a.length_field_length
    e = Uint16(1)
    a.data.append(e)
    assert len(a) == a.length_field_length + len(Uint16())
    del a.data[0]
    assert len(a) == a.length_field_length

    a.length_field_length = 2
    # Except only the length field with 2 bytes to be serialized and no data
    assert bytes.fromhex("0000") == a.serialize()

    a.length_field_length = 4
    # Except only the length field with 4 bytes to be serialized and no data
    assert bytes.fromhex("00000000") == a.serialize()

    a.data.append(Uint16(1))
    assert bytes.fromhex("000000020001") == a.serialize()

    a.data.append(Uint16(4))
    assert len(a) == 2 * len(Uint16()) + a.length_field_length
    assert bytes.fromhex("0000000400010004") == a.serialize()

    b = SomeIpDynamicSizeArray(Uint16)
    b.length_field_length = 4
    b = b.deserialize(bytes.fromhex("0000000400010004"))
    assert b == a
    assert b.data[0] == Uint16(1)
    assert b.data[1] == Uint16(4)
    assert len(b) == 2 * len(Uint16()) + a.length_field_length


def test_someip_fixed_size_string():
    a = SomeIpFixedSizeString(4)
    assert a.size == 4
    assert a.data == ""
    assert a.encoding == "utf-8"
    assert len(a) == 4 + 3

    with pytest.raises(ValueError):
        a.data = "Hello World"
    with pytest.raises(ValueError):
        a.encoding = "Hello World"

    a.data = "He"
    a.encoding = "utf-8"
    assert len(a) == 4 + 3

    b = SomeIpFixedSizeString(4)
    b.data = "He"
    assert a == b

    assert bytes.fromhex("EF BB BF 48 65 00 00") == a.serialize()

    a.encoding = "utf-16le"
    assert bytes.fromhex("FF FE 48 00 65 00 00 00 00 00") == a.serialize()

    a.encoding = "utf-16be"
    assert bytes.fromhex("FE FF 00 48 00 65 00 00 00 00") == a.serialize()

    b = SomeIpFixedSizeString(4)
    b.encoding = "utf-16be"
    b = b.deserialize(bytes.fromhex("EF BB BF 48 65 00 00 00 00"))
    assert b.encoding == "utf-8"
    assert b.data == "He"
    assert len(b) == 7

    b = b.deserialize(bytes.fromhex("FF FE 48 00 65 00 00 00 00 00"))
    assert b.encoding == "utf-16le"
    assert b.data == "He"
    assert len(b) == 10

    b = b.deserialize(bytes.fromhex("FE FF 00 48 00 65 00 00 00 00"))
    assert b.encoding == "utf-16be"
    assert b.data == "He"
    assert len(b) == 10


def test_someip_dynamic_size_string():
    a = SomeIpDynamicSizeString()
    assert a.data == ""
    assert a.encoding == "utf-8"
    assert a.length_field_length == 4
    assert a._length_field_value == 4
    assert len(a) == 4 + 3 + 1

    with pytest.raises(ValueError):
        a.encoding = "Hello World"
    with pytest.raises(ValueError):
        a.length_field_length = 0

    a.data = "He"
    a.encoding = "utf-8"
    assert len(a) == 4 + 3 + 2 + 1
    # 4 bytes for the length field, 3 bytes BOM and 3 bytes for the data including '\0'
    assert bytes.fromhex("00 00 00 06 EF BB BF 48 65 00") == a.serialize()

    a.encoding = "utf-16le"
    assert len(a) == 4 + 2 + 2 * 2 + 2
    assert bytes.fromhex("00 00 00 08 FF FE 48 00 65 00 00 00") == a.serialize()

    a.encoding = "utf-16be"
    assert len(a) == 4 + 2 + 2 * 2 + 2
    assert bytes.fromhex("00 00 00 08 FE FF 00 48 00 65 00 00") == a.serialize()

    a.encoding = "utf-8"
    a.length_field_length = 2
    assert len(a) == 2 + 3 + 2 + 1
    # 2 bytes for the length field, 3 bytes BOM and 3 bytes for the data including '\0'
    assert bytes.fromhex("00 06 EF BB BF 48 65 00") == a.serialize()

    b = SomeIpDynamicSizeString()
    b.encoding = "utf-16be"
    b.length_field_length = 4
    b = b.deserialize(bytes.fromhex("00 00 00 09 EF BB BF 48 65 00 00 00 00"))
    assert len(b) == 13
    assert b.encoding == "utf-8"
    assert b.data == "He"

    b = b.deserialize(bytes.fromhex("00 00 00 0A FF FE 48 00 65 00 00 00 00 00"))
    assert b.encoding == "utf-16le"
    assert b.data == "He"

    b = b.deserialize(bytes.fromhex("00 00 00 0A FE FF 00 48 00 65 00 00 00 00"))
    assert b.encoding == "utf-16be"
    assert b.data == "He"

    b.length_field_length = 2
    b = b.deserialize(bytes.fromhex("00 0A FE FF 00 48 00 65 00 00 00 00"))
    assert b.encoding == "utf-16be"
    assert b.data == "He"


@dataclass
class MsgWithStrings(SomeIpPayload):
    a: Uint16
    b: SomeIpDynamicSizeString
    d: Uint32

    def __init__(self):
        self.a = Uint16(10)
        self.b = SomeIpDynamicSizeString("123")
        self.c = Uint32(5)


def test_struct_with_dynamic_sized_types():
    a = MsgWithStrings()
    assert len(a) == len(Uint16()) + len(SomeIpDynamicSizeString("123")) + len(Uint32())
    assert len(a) == 2 + 4 + 3 + 3 + 1 + 4
    assert len(a.serialize()) == len(a)

    assert (
        bytes.fromhex("00 0A 00 00 00 07 EF BB BF 31 32 33 00 00 00 00 05")
        == a.serialize()
    )

    a.b = SomeIpDynamicSizeString("1234")
    assert len(a) == 2 + 4 + 3 + 3 + 1 + 4 + 1
    assert (
        bytes.fromhex("00 0A 00 00 00 08 EF BB BF 31 32 33 34 00 00 00 00 05")
        == a.serialize()
    )

    b = MsgWithStrings().deserialize(
        bytes.fromhex("00 0A 00 00 00 08 EF BB BF 31 32 33 34 00 00 00 00 05")
    )
    assert b.a == Uint16(10)
    assert b.b == SomeIpDynamicSizeString("1234")
    assert b.c == Uint32(5)


@dataclass
class MsgWithTwoStrings(SomeIpPayload):
    a: Uint16
    b: SomeIpDynamicSizeString
    c: SomeIpDynamicSizeString
    d: Uint32

    def __init__(self):
        self.a = Uint16(10)
        self.b = SomeIpDynamicSizeString("123")
        self.c = SomeIpDynamicSizeString("4321")
        self.c.encoding = "utf-16be"
        self.d = Uint32(5)


def test_struct_with_dynamic_sized_types():
    a = MsgWithTwoStrings()
    assert len(a) == 2 + 4 + 3 + 3 + 1 + 4 + 2 + 4 * 2 + 2 + 4
    assert len(a.serialize()) == len(a)

    assert (
        bytes.fromhex(
            "00 0A 00 00 00 07 EF BB BF 31 32 33 00 00 00 00 0C FE FF 00 34 00 33 00 32 00 31 00 00 00 00 00 05"
        )
        == a.serialize()
    )

    b = MsgWithTwoStrings().deserialize(
        bytes.fromhex(
            "00 0A 00 00 00 07 EF BB BF 31 32 33 00 00 00 00 0C FE FF 00 34 00 33 00 32 00 31 00 00 00 00 00 05"
        )
    )
    assert b.a == Uint16(10)
    assert b.b == SomeIpDynamicSizeString("123")

    c = SomeIpDynamicSizeString("4321")
    c.encoding = "utf-16be"
    assert b.c == c
    assert b.d == Uint32(5)


@dataclass
class MsgWithDynamicArrays(SomeIpPayload):
    a: Uint16
    b: SomeIpDynamicSizeArray[Uint8]
    c: Uint32
    d: SomeIpDynamicSizeArray[Uint16]

    def __init__(self):
        self.a = Uint16(10)
        self.b = SomeIpDynamicSizeArray(Uint8)
        self.c = Uint32(5)
        self.d = SomeIpDynamicSizeArray(Uint16)


def test_struct_with_dynamic_arrays():
    a = MsgWithDynamicArrays()
    assert len(a) == 2 + 4 + 4 + 4
    assert len(a.serialize()) == len(a)
    assert bytes.fromhex("00 0A 00 00 00 00 00 00 00 05 00 00 00 00") == a.serialize()

    a.b.data.append(Uint8(1))
    assert len(a.serialize()) == len(a)
    assert (
        bytes.fromhex("00 0A 00 00 00 01 01 00 00 00 05 00 00 00 00") == a.serialize()
    )

    b = MsgWithDynamicArrays().deserialize(
        bytes.fromhex("00 0A 00 00 00 01 01 00 00 00 05 00 00 00 00")
    )

    assert b.a == Uint16(10)
    assert b.b.data[0] == Uint8(1)
    assert b.c == Uint32(5)
    assert len(b.d) == b.d.length_field_length



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from collections.abc import Callable
import logging
from typing import Tuple
from someipy._internal._common.endpoint import Endpoint
from someipy._internal.someip_endpoint import (
    SomeipEndpoint,
    TCPClientSomeipEndpoint,
    TCPSomeipEndpoint,
    UDPSomeipEndpoint,
)
from someipy._internal.someip_message import SomeIpMessage
from someipy._internal.tcp_client_manager import TcpClientManager, TcpClientProtocol
from someipy._internal.transport_layer_protocol import TransportLayerProtocol
from someipy._internal.utils import create_udp_socket


class SomeipEndpointFactory:

    @staticmethod
    async def create_server_endpoint(
        endpoint: Endpoint,
        protocol: TransportLayerProtocol,
        someip_callback: Callable[
            [SomeIpMessage, Tuple[str, int], Tuple[str, int], TransportLayerProtocol],
            None,
        ],
    ) -> SomeipEndpoint:

        if protocol == TransportLayerProtocol.UDP:
            loop = asyncio.get_running_loop()
            rcv_socket = create_udp_socket(str(endpoint.ip), endpoint.port)

            _, udp_endpoint = await loop.create_datagram_endpoint(
                lambda: UDPSomeipEndpoint(str(endpoint.ip), endpoint.port),
                sock=rcv_socket,
            )

            udp_endpoint.set_someip_callback(someip_callback)

            return udp_endpoint
        else:
            tcp_client_manager = TcpClientManager(str(endpoint.ip), endpoint.port)
            loop = asyncio.get_running_loop()
            server = await loop.create_server(
                lambda: TcpClientProtocol(client_manager=tcp_client_manager),
                str(endpoint.ip),
                endpoint.port,
            )
            tcp_someip_endpoint = TCPSomeipEndpoint(
                server, tcp_client_manager, str(endpoint.ip), endpoint.port
            )

            tcp_someip_endpoint.set_someip_callback(someip_callback)

            return tcp_someip_endpoint

    @staticmethod
    async def create_udp_client_endpoint(
        dst_endpoint: Endpoint,
        src_endpoint: Endpoint,
        someip_message_callback: Callable[[SomeIpMessage], None],
        logger: logging.Logger = None,
    ) -> SomeipEndpoint:
        udp_endpoint = await SomeipEndpointFactory.create_server_endpoint(
            src_endpoint,
            TransportLayerProtocol.UDP,
            someip_message_callback,
        )
        return udp_endpoint

    @staticmethod
    def create_tcp_client_endpoint(
        dst_endpoint: Endpoint,
        src_endpoint: Endpoint,
        someip_message_callback: Callable[[SomeIpMessage], None],
        logger: logging.Logger = None,
    ) -> TCPClientSomeipEndpoint:
        tcp_endpoint = TCPClientSomeipEndpoint(
            str(dst_endpoint.ip),
            dst_endpoint.port,
            str(src_endpoint.ip),
            src_endpoint.port,
            logger,
        )
        tcp_endpoint.set_someip_callback(someip_message_callback)
        return tcp_endpoint



from ._internal.transport_layer_protocol import TransportLayerProtocol  # noqa: F401
from .service import Service, ServiceBuilder, EventGroup, Method, Event  # noqa: F401
from .server_service_instance import ServerServiceInstance  # noqa: F401
from .client_service_instance import ClientServiceInstance  # noqa: F401
from ._internal.someip_message import SomeIpMessage  # noqa: F401
from ._internal.method_result import MethodResult  # noqa: F401
from ._internal.return_codes import ReturnCode  # noqa: F401
from ._internal.message_types import MessageType  # noqa: F401
from ._internal._daemon.someipy_daemon_client import (
    connect_to_someipy_daemon,
)  # noqa: F401
